import marshal
import mmap
import os
import os.path
import sys
import logging
import threading
import queue
# from profilehooks import profile
from cog.cache import Cache
import xxhash

RECORD_SEP = b'\xFD'
UNIT_SEP = b'\xAC'


class TableMeta:
    __slots__ = ('name', 'namespace', 'db_instance_id', 'column_mode')

    def __init__(self, name, namespace, db_instance_id, column_mode):
        self.name = name
        self.namespace = namespace
        self.db_instance_id = db_instance_id
        self.column_mode = column_mode


class Table:

    def __init__(self, name, namespace, db_instance_id, config, column_mode=False, shared_cache=None,
                 flush_interval=1):
        self.logger = logging.getLogger('table')
        self.config = config
        self.shared_cache = shared_cache
        self.flush_interval = flush_interval
        self.table_meta = TableMeta(name, namespace, db_instance_id, column_mode)
        self.indexer = self.__create_indexer()
        self.store = self.__create_store(shared_cache)

    def __create_indexer(self):
        return Indexer(self.table_meta, self.config, self.logger)

    def __create_store(self, shared_cache):
        return Store(self.table_meta, self.config, self.logger, shared_cache=shared_cache,
                     flush_interval=self.flush_interval)

    def sync(self):
        """Force flush pending writes to disk."""
        self.store.sync()

    def close(self):
        self.indexer.close()
        self.store.close()
        self.logger.info("closed table: " + self.table_meta.name)


class Record:
    '''
    Record is the basic unit of storage in cog.
    value_type: s - string, l - list, u - set
    '''
    __slots__ = ('key', 'value', 'tombstone', 'store_position', 'key_link', 'value_link', 'value_type')
    
    RECORD_LINK_LEN = 16
    RECORD_LINK_NULL = -1
    VALUE_LINK_NULL = -1

    def __init__(self, key, value, tombstone='0', store_position=None, value_type="s", key_link=-1, value_link=-1):
        self.key = key
        self.value = value
        self.tombstone = tombstone
        self.store_position = store_position
        self.key_link = key_link
        self.value_link = value_link
        self.value_type = value_type

    def set_store_position(self, pos):
        if type(pos) is not int:
            raise ValueError("store position must be int but provided : " + str(pos))
        self.store_position = pos

    def set_key_link(self, pos):
        self.key_link = pos

    def set_value_link(self, pos):
        self.value_link = pos

    def set_value(self, value):
        self.value = value

    def is_equal_val(self, other_record):
        return self.key == other_record.key and self.value == other_record.value

    def get_kv_tuple(self):
        return self.key, self.value

    def serialize(self):
        return marshal.dumps((self.key, self.value))

    def marshal(self):
        key_link_bytes = str(self.key_link).encode().rjust(Record.RECORD_LINK_LEN)
        serialized = self.serialize()
        m_record = key_link_bytes \
                   + self.tombstone.encode() \
                   + self.value_type.encode() \
                   + str(len(serialized)).encode() \
                   + UNIT_SEP \
                   + serialized
        if self.value_type == "l" or self.value_type == "u":
            if self.value_link is not None:
                m_record += str(self.value_link).encode()
        m_record += RECORD_SEP
        return m_record

    def is_empty(self):
        return self.key is None and self.value is None

    def __str__(self):
        return "key: {}, value: {}, tombstone: {}, store_position: {}, key_link: {}, value_link: {}, value_type: {}".format(
            self.key, self.value, self.tombstone, self.store_position, self.key_link, self.value_link, self.value_type)

    @classmethod
    def __read_until(cls, start, sbytes, separtor=UNIT_SEP):
        buff = b''
        i = 0  # default
        for i in range(start, len(sbytes)):
            s_byte = sbytes[i: i + 1]
            if s_byte == separtor:
                break
            buff += s_byte
        return buff, i

    @classmethod
    def unmarshal(cls, store_bytes):
        """reads from bytes and creates object
        """
        base_pos = 0
        key_link = int(store_bytes[base_pos: base_pos + Record.RECORD_LINK_LEN])
        next_base_pos = Record.RECORD_LINK_LEN
        tombstone = store_bytes[next_base_pos:next_base_pos + 1].decode()
        value_type = store_bytes[next_base_pos + 1: next_base_pos + 2].decode()
        value_len, end_pos = cls.__read_until(next_base_pos + 2, store_bytes)
        value_len = int(value_len.decode())
        value = store_bytes[end_pos + 1: end_pos + 1 + value_len]
        record = marshal.loads(value)

        value_link = Record.VALUE_LINK_NULL
        if value_type == 'l' or value_type == 'u':
            value_link, end_pos = cls.__read_until(end_pos + value_len + 1, store_bytes, RECORD_SEP)
            value_link = int(value_link.decode())
        return cls(record[0], record[1], tombstone, store_position=None, value_type=value_type, key_link=key_link,
                   value_link=value_link)

    @classmethod
    def __load_value(cls, store_pointer, val_list, store):
        """loads value from the store"""
        while store_pointer != Record.VALUE_LINK_NULL:
            rec = Record.unmarshal(store.read(store_pointer))
            if rec.value_type == 'l':
                val_list.append(rec.value)
            else:
                val_list.add(rec.value)
            store_pointer = rec.value_link
        return val_list

    @classmethod
    # @profile
    def load_from_store(cls, position: int, store):
        record = cls.unmarshal(store.read(position))
        values = None
        if record.value_type == 'l':
            values = cls.__load_value(record.value_link, [record.value], store)
        elif record.value_type == 'u':
            values = cls.__load_value(record.value_link, {record.value}, store)
        if values is not None:
            record.set_value(values)
        return record


class Index:

    def __init__(self, table_meta, config, logger, index_id=0):
        self.logger = logging.getLogger('index')
        self.table = table_meta
        self.config = config
        self.name = self.config.cog_index(table_meta.namespace, table_meta.name, table_meta.db_instance_id, index_id)
        self.empty_block = '-1'.zfill(self.config.INDEX_BLOCK_LEN).encode()
        if not os.path.exists(self.name):
            self.logger.info("creating index...")
            f = open(self.name, 'wb+')
            i = 0
            e_blocks = []
            while i < config.INDEX_CAPACITY:
                e_blocks.append(self.empty_block)
                i += 1
            f.write(b''.join(e_blocks))
            self.file_limit = f.tell()
            f.close()
            self.logger.info("new index with capacity" + str(config.INDEX_CAPACITY) + "created: " + self.name)
        else:
            self.logger.info("Index: "+self.name+" already exists.")

        self.db = open(self.name, 'r+b')
        self.db_mem = mmap.mmap(self.db.fileno(), 0)
        self._closed = False

    def close(self):
        if self._closed:
            return
        self._closed = True
        self.db_mem.flush()
        self.db_mem.close()
        self.db.close()

    def get_index_key(self, int_store_position):
        return str(int_store_position).encode().rjust(self.config.INDEX_BLOCK_LEN)

    # @profile
    def put(self, key, store_position, store):
        """
        key chain
        :param key:
        :param store_position:
        :param store:
        :return:
        """

        """
        k5 -> k4 -> k3 -> k2 -> k1
        add: k6
        k6 -> k5 -> k4 -> k3 -> k2 -> k1
        add/update: k4
        1. k4 -> k6 -> k5 -> k4 -> k3 -> k2 -> k1
        2. k4 -> k6 -> k5 -> k3 -> k2 -> k1

        """
        orig_position, orig_hash = self.get_index(key)
        index_value = self.db_mem[orig_position: orig_position + self.config.INDEX_BLOCK_LEN]
        self.logger.debug('writing : ' + str(key) + ' current data at store position: ' + str(index_value))
        if index_value == self.empty_block:
            # First record in the key bucket, point next link to null
            store.update_record_link_inplace(store_position, Record.RECORD_LINK_NULL)
            key_link = store_position
            # write the store position to the index
            self.db_mem[orig_position: orig_position + self.config.INDEX_BLOCK_LEN] = self.get_index_key(store_position)
        else:
            # there are records in the bucket
            # read existing record from the store - use unmarshal, not load_from_store (O(1) vs O(n))
            existing_record = Record.unmarshal(store.read(int(index_value)))
            existing_record.set_store_position(int(index_value))

            if existing_record.key == key:
                # the record at the top of the bucket has the same key, update the record in place.
                # a new entry has been made to the store, update pre and next links.
                store.update_record_link_inplace(store_position, int(existing_record.key_link))
                key_link = int(existing_record.key_link)
            else:
                # this is hash collision.
                # the record at the top of the bucket has a different key, add new record to the top of the bucket.
                # set next link to the record at the top of the bucket, prev is null by default, no need to set.
                store.update_record_link_inplace(store_position, existing_record.store_position)
                key_link = existing_record.store_position

                # check if this record exists in the bucket, if yes remove pointer.
                prev_record = None
                while existing_record.key_link != Record.RECORD_LINK_NULL:
                    # Use unmarshal (O(1)) instead of load_from_store (O(n))
                    existing_record = Record.unmarshal(store.read(existing_record.key_link))
                    existing_record.set_store_position(existing_record.key_link)
                    if existing_record.key == key and prev_record is not None:
                        """
                        if same key found in bucket, update previous record in chain to point to key_link of this record
                        prev_rec -> current rec.key_link
                        curr_rec will not be linked in the bucket anymore.
                        """
                        # update in place the key link pointer of pervios record, ! need to add fixed length padding.
                        store.update_record_link_inplace(prev_record.store_position, existing_record.key_link)
                        key_link = existing_record.key_link
                    prev_record = existing_record

        self.db_mem[orig_position: orig_position + self.config.INDEX_BLOCK_LEN] = self.get_index_key(store_position)
        return key_link

    def get_index(self, key):
        num = cog_hash(key, self.config.INDEX_CAPACITY) % ((sys.maxsize + 1) * 2)
        self.logger.debug("hash for: " + key + " : " + str(num))
        # there may be diff when using mem slice vs write (+1 needed)
        index = (self.config.INDEX_BLOCK_LEN *
                 (max((num % self.config.INDEX_CAPACITY) - 1, 0)))
        self.logger.debug("offset : " + key + " : " + str(index))
        return index, num

    def cog_hash(self, string):
        return xxhash.xxh32(string, seed=2).intdigest() % self.config.INDEX_CAPACITY

    # @profile
    def get(self, key, store):
        self.logger.debug("GET: Reading index: " + self.name)
        index_position, raw_hash = self.get_index(key)
        data_at_index_position = self.db_mem[index_position:index_position + self.config.INDEX_BLOCK_LEN]
        if data_at_index_position == self.empty_block:
            return None
        data_at_index_position = int(data_at_index_position)
        record = Record.load_from_store(data_at_index_position, store)
        record.set_store_position(data_at_index_position)
        self.logger.debug("read record " + str(record))

        if record.key == key:
            return record
        else:
            while record.key_link != Record.RECORD_LINK_NULL:
                self.logger.debug("record.key_link: " + str(record.key_link))
                record = Record.load_from_store(record.key_link, store)
                record.set_store_position(record.key_link)
                if record.key == key:
                    return record
        return None

    def get_head_only(self, key, store):
        """
        Get only the head record without traversing the value chain.
        This is O(1) compared to get() which is O(n) for multi-value keys.
        
        Returns: (record, store_position) or (None, None)
        """
        index_position, raw_hash = self.get_index(key)
        data_at_index_position = self.db_mem[index_position:index_position + self.config.INDEX_BLOCK_LEN]
        if data_at_index_position == self.empty_block:
            return None, None
        store_position = int(data_at_index_position)
        # Only unmarshal, don't load value chain
        record = Record.unmarshal(store.read(store_position))
        record.set_store_position(store_position)
        
        if record.key == key:
            return record, store_position
        else:
            # Hash collision - follow key_link chain
            while record.key_link != Record.RECORD_LINK_NULL:
                store_position = record.key_link
                record = Record.unmarshal(store.read(store_position))
                record.set_store_position(store_position)
                if record.key == key:
                    return record, store_position
        return None, None

    '''
        Iterates through all records in the index, following key_link chains for hash collisions.
    '''

    def scanner(self, store):
        scan_cursor = 0
        while True:
            data_at_position = self.db_mem[scan_cursor:scan_cursor + self.config.INDEX_BLOCK_LEN]
            if len(data_at_position) == 0:  # EOF index
                self.logger.info("Index EOF reached! Scan terminated.")
                return
            if data_at_position == self.empty_block:
                scan_cursor += self.config.INDEX_BLOCK_LEN
                self.logger.debug("GET: skipping empty block during iteration.")
                continue
            
            # Load head record and follow key_link chain to get all records in this bucket
            store_position = int(data_at_position)
            while store_position != Record.RECORD_LINK_NULL:
                record = Record.load_from_store(store_position, store)
                if record is None:  # EOF store
                    self.logger.error("Store EOF reached! Iteration terminated.")
                    return
                yield Record(record.key, record.value, record.tombstone)
                store_position = record.key_link
            
            scan_cursor += self.config.INDEX_BLOCK_LEN

    def delete(self, key, store):
        """
               k5 -> k4 -> k3 -> k2 -> k1
               del: k3
               k6 -> k5 -> k4 -> k2 -> k1

        """
        self.logger.debug("GET: Reading index: " + self.name)
        index_position, raw_hash = self.get_index(key)

        data_at_index_position = self.db_mem[index_position:index_position + self.config.INDEX_BLOCK_LEN]
        if data_at_index_position == self.empty_block:
            return False

        data_at_index_position = int(data_at_index_position)
        record = Record.load_from_store(data_at_index_position, store)
        record.set_store_position(data_at_index_position)
        self.logger.debug("read record " + str(record))
        if record.key == key:
            """delete bucket => map hash table to empty block, or point to next in chain"""
            if record.key_link != Record.RECORD_LINK_NULL:
                # Point index to next record in chain
                self.db_mem[index_position:index_position + self.config.INDEX_BLOCK_LEN] = self.get_index_key(record.key_link)
            else:
                # No more records in chain, clear the bucket
                self.db_mem[index_position:index_position + self.config.INDEX_BLOCK_LEN] = self.empty_block
            return True
        else:
            """search bucket"""
            prev_record = record  # Initialize to the head record
            while record.key_link != Record.RECORD_LINK_NULL:
                next_record = Record.load_from_store(record.key_link, store)
                next_record.set_store_position(record.key_link)
                if next_record.key == key:
                    """
                    if same key found in bucket, update previous record in chain to point to key_link of this record
                    prev_rec -> current rec.key_link
                    curr_rec will not be linked in the bucket anymore.
                    """
                    # update in place the key link pointer of previous record
                    store.update_record_link_inplace(prev_record.store_position, next_record.key_link)
                    return True
                prev_record = next_record
                record = next_record
        return False

    def flush(self):
        self.db_mem.flush()


class Store:
    """
    Store manages persistence of records to disk with configurable flush behavior.
    
    Args:
        flush_interval: Number of writes before auto-flush. 
                       1 = flush every write (safest, default)
                       0 = manual flush only (fastest, use sync())
                       N>1 = flush every N writes with async background thread
    """

    def __init__(self, tablemeta, config, logger, caching_enabled=True, shared_cache=None, 
                 flush_interval=1):
        self.caching_enabled = caching_enabled
        self.batch_mode = False  # When True, defers flush() until end_batch()
        self.logger = logging.getLogger('store')
        self.tablemeta = tablemeta
        self.config = config
        self.flush_interval = flush_interval
        self.write_count = 0
        self._closed = False
        self.empty_block = '-1'.zfill(self.config.INDEX_BLOCK_LEN).encode()
        self.store = self.config.cog_store(
            tablemeta.namespace, tablemeta.name, tablemeta.db_instance_id)
        self.store_cache = Cache(self.store, shared_cache)
        temp = open(self.store, 'a')  # create if not exist
        temp.close()
        self.store_file = open(self.store, 'rb+')
        
        # Thread safety
        self._lock = threading.Lock()
        
        # Auto-enable async flush when interval > 1
        self._use_async = flush_interval > 1
        if self._use_async:
            self._flush_queue = queue.Queue()
            self._flush_thread = threading.Thread(target=self._flush_worker, daemon=True)
            self._flush_thread.start()
            self._shutdown = False
        
        logger.info(f"Store init: {self.store} (flush_interval={flush_interval})")

    def _flush_worker(self):
        """Background thread that processes flush requests."""
        while True:
            try:
                # Wait for flush signal (blocks until item available)
                item = self._flush_queue.get(timeout=1.0)
                if item == "SHUTDOWN":
                    # Drain remaining items and shutdown
                    while not self._flush_queue.empty():
                        try:
                            self._flush_queue.get_nowait()
                            self._flush_queue.task_done()
                        except queue.Empty:
                            break
                    self._flush_queue.task_done()
                    break
                # Perform actual flush (check if not closed)
                if not self._closed:
                    with self._lock:
                        if not self._closed:
                            self.store_file.flush()
                self._flush_queue.task_done()
            except queue.Empty:
                # Timeout - check if we should continue
                if getattr(self, '_shutdown', False):
                    break
                continue

    def _request_flush(self):
        """Request a flush - async if interval > 1, sync otherwise."""
        if self._closed:
            return
        if self._use_async:
            self._flush_queue.put("FLUSH")
        else:
            self.store_file.flush()

    def _handle_write_flush(self):
        """Increment write count and trigger flush if threshold reached."""
        if not self.batch_mode:
            self.write_count += 1
            if self.flush_interval > 0 and self.write_count >= self.flush_interval:
                self._request_flush()
                self.write_count = 0

    def sync(self):
        """
        Force flush all pending writes to disk.
        Blocks until flush is complete.
        """
        if self._closed:
            return
        with self._lock:
            if not self._closed:
                self.store_file.flush()
        if self._use_async:
            # Wait for async queue to drain
            self._flush_queue.join()

    def close(self):
        """Close the store, ensuring all data is flushed."""
        if self._closed:
            return
            
        # Mark as closed first
        self._closed = True
        
        if self._use_async:
            self._shutdown = True
            self._flush_queue.put("SHUTDOWN")
            self._flush_thread.join(timeout=5.0)
        
        with self._lock:
            try:
                self.store_file.flush()
                self.store_file.close()
            except ValueError:
                pass  # File already closed

    def begin_batch(self):
        """
        Enable batch mode - defers flush() until end_batch() is called.
        Use this when inserting many records for significantly better performance.
        """
        self.batch_mode = True

    def end_batch(self):
        """
        End batch mode and flush all pending writes to disk.
        """
        with self._lock:
            self.store_file.flush()
        self.batch_mode = False

    def save(self, record):
        """
        Store data with configurable flush behavior.
        """
        with self._lock:
            self.store_file.seek(0, 2)
            store_position = self.store_file.tell()
            record.set_store_position(store_position)
            marshalled_record = record.marshal()
            self.store_file.write(marshalled_record)
            
            if self.caching_enabled:
                self.store_cache.put(store_position, marshalled_record)
            
            # Handle flush based on interval
            self._handle_write_flush()
        
        return store_position

    def update_record_link_inplace(self, start_pos, int_value):
        """updates record link in store file in place"""
        if type(int_value) is not int:
            raise ValueError("store position must be int but provided : " + str(start_pos))

        byte_value = str(int_value).encode().rjust(Record.RECORD_LINK_LEN)
        self.logger.debug('update_record_link_inplace: ' + str(byte_value))
        
        with self._lock:
            self.store_file.seek(start_pos)
            self.store_file.write(byte_value)

            if self.caching_enabled:
                self.store_cache.partial_update_from_zero_index(start_pos, byte_value)
            
            self._handle_write_flush()

    # @profile
    def read(self, position):
        self.logger.debug("store read request at position: " + str(position))
        if self.caching_enabled:
            cached_record = self.store_cache.get(position)
            if cached_record is not None:
                return cached_record

        self.store_file.seek(position)
        record = self.__read_record()

        if self.caching_enabled:
            self.store_cache.put(position, record)

        return record

    def __read_record(self):
        """Read a single record using structured/length-aware parsing.

        Record layout (written by Record.marshal):
            [key_link: 16 bytes]
            [tombstone: 1 byte]
            [value_type: 1 byte]
            [value_len digits: variable ASCII]
            [UNIT_SEP: 1 byte]
            [serialized payload: exactly value_len bytes]
            [value_link digits: variable ASCII, only if value_type in ('l','u')]
            [RECORD_SEP: 1 byte]
        """
        # --- fixed-size header: key_link(16) + tombstone(1) + value_type(1) = 18 bytes
        header = self.__read_exactly(18)
        if header is None or len(header) < 18:
            return None

        value_type = chr(header[17])       # 's', 'l', or 'u'

        # --- variable-length value_len digits terminated by UNIT_SEP
        len_buf = b''
        while True:
            b = self.__read_exactly(1)
            if b is None:
                return None
            if b == UNIT_SEP:
                break
            len_buf += b
        value_len = int(len_buf.decode())

        # --- serialized payload: exactly value_len bytes
        payload = self.__read_exactly(value_len)
        if payload is None or len(payload) < value_len:
            return None

        # --- optional value_link + RECORD_SEP terminator
        tail = b''
        if value_type in ('l', 'u'):
            # value_link digits (ASCII) terminated by RECORD_SEP
            while True:
                b = self.__read_exactly(1)
                if b is None:
                    break
                tail += b
                if b == RECORD_SEP:
                    break
        else:
            # just the RECORD_SEP terminator
            sep = self.__read_exactly(1)
            if sep is not None:
                tail = sep

        record = header + len_buf + UNIT_SEP + payload + tail
        self.logger.debug("store __read_record: " + str(record))
        return record

    def __read_exactly(self, n):
        """Read exactly *n* bytes from the store file, or return None on EOF."""
        data = self.store_file.read(n)
        if len(data) == 0:
            return None
        while len(data) < n:
            chunk = self.store_file.read(n - len(data))
            if len(chunk) == 0:
                return data          # partial read (EOF mid-record)
            data += chunk
        return data


class Indexer:
    '''
    Manages indexes. Creates new index when an index is full.
    Searches all indexes for get requests.
    Provides same get/put/del method as single index but over multuple files.
    '''

    def __init__(self, tablemeta, config, logger):
        self.tablemeta = tablemeta
        self.config = config
        self.logger = logging.getLogger('indexer')
        self.index_list = []  # future range index.
        self.index_id = 0
        self.load_indexes()
        # if no index currenlty exist, create new live index.
        if len(self.index_list) == 0:
            self.index_list.append(Index(tablemeta, config, logger, self.index_id))
            self.live_index = self.index_list[self.index_id]

    def close(self):
        for idx in self.index_list:
            idx.close()

    def load_indexes(self):
        for f in os.listdir(self.config.cog_data_dir(self.tablemeta.namespace)):
            if self.config.INDEX in f:
                if self.tablemeta.name == self.config.get_table_name(f):
                    self.logger.info("loading index file: " + f)
                    id = self.config.index_id(f)
                    index = Index(self.tablemeta, self.config, self.logger, id)
                    self.index_list.append(index)
                    # make the latest index the live index.
                    if id >= self.index_id:
                        self.index_id = id
                        self.live_index = index

    def put(self, key, store_position, store):
        resp = self.live_index.put(key, store_position, store)
        self.logger.debug("Key: " + key + " indexed in: " + self.live_index.name)
        return resp

    # @profile
    def get(self, key, store):
        idx = self.index_list[0]  # only one index file.
        return idx.get(key, store)

    def get_head_only(self, key, store):
        """Get head record only, O(1) - doesn't traverse value chain."""
        idx = self.index_list[0]
        return idx.get_head_only(key, store)

    def scanner(self, store):
        for idx in self.index_list:
            self.logger.debug("SCAN: index: " + idx.name)
            for r in idx.scanner(store):
                yield r

    def delete(self, key, store):
        for idx in self.index_list:
            if idx.delete(key, store):
                return True
            else:
                return False

def cog_hash(string, index_capacity):
    return xxhash.xxh32(string, seed=2).intdigest() % index_capacity



from cog.core import Record
from cog.database import Cog
from cog import config
import json
import unittest
import os
import shutil


def qfilter(jsn):
    d = json.loads(jsn[1])
    return d["firstname"]


DIR_NAME = "TestDB"


class TestDB(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if not os.path.exists("/tmp/"+DIR_NAME+"/"):
            os.mkdir("/tmp/" + DIR_NAME + "/")
            os.mkdir("/tmp/"+DIR_NAME+"/test_table/")

        config.COG_HOME = DIR_NAME

    def test_db(self):
        data = Record('user100','{"firstname":"Hari","lastname":"seldon"}')
        cogdb = Cog()
        cogdb.create_or_load_namespace("test")
        cogdb.create_table("db_test", "test")
        cogdb.put(data)
        scanner = cogdb.scanner()
        res = None
        for r in scanner:
            res = r
        print(res)
        self.assertTrue(data.is_equal_val(res))
        cogdb.close()

    def test_list_tables(self):
        cogdb = Cog()
        cogdb.create_or_load_namespace("test_ns")
        cogdb.create_table("table1", "test_ns")
        cogdb.create_table("table2", "test_ns")
        cogdb.create_table("table3", "test_ns")
        self.assertEqual(set(cogdb.list_tables()), {'table2', 'table3', 'table1'})
        cogdb.close()


    @classmethod
    def tearDownClass(cls):
        shutil.rmtree("/tmp/"+DIR_NAME)
        print("*** deleted test data.")


if __name__ == '__main__':
    unittest.main()



from cog.core import Record
from cog.database import Cog
from cog import cog
import unittest
import os
import shutil

DIR_NAME = "TestLib"


class TestLib(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if not os.path.exists("/tmp/"+DIR_NAME+"/"):
            os.mkdir("/tmp/" + DIR_NAME + "/")
            os.mkdir("/tmp/"+DIR_NAME+"/test/")

    def test_cog_function(self):
        """Test the cog() function returns expected message."""
        result = cog()
        self.assertEqual(result, "Cog is alive.")

    def test_db(self):
        data = Record('testKey','testVal')
        cogdb = Cog()
        cogdb.create_or_load_namespace("test")
        cogdb.create_table("db_test", "test")
        cogdb.put(data)
        self.assertTrue(cogdb.get("testKey").is_equal_val(Record('testKey', 'testVal')))
        cogdb.close()

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree("/tmp/"+DIR_NAME)
        print("*** deleted test data.")


if __name__ == '__main__':
    unittest.main()


"""
Tests for JSON operations (putj, updatej) to ensure correct graph construction
and update semantics for nested objects, arrays, and mixed structures.
"""

import pytest
import os
import shutil
from cog.torque import Graph
from cog import config


COG_HOME = "test_json_ops_isolated_home"


@pytest.fixture
def clean_graph():
    """Create a fresh graph for each test with complete isolation."""
    import uuid
    unique_dir = COG_HOME + "_" + str(uuid.uuid4())[:8]
    full_path = "/tmp/" + unique_dir
    
    if os.path.exists(full_path):
        shutil.rmtree(full_path)
    os.makedirs(full_path)
    config.CUSTOM_COG_DB_PATH = full_path
    
    g = Graph("test_graph")
    yield g
    g.close()
    if os.path.exists(full_path):
        shutil.rmtree(full_path)


class TestPutJson:
    """Tests for putj - inserting JSON objects into the graph."""

    def test_simple_object(self, clean_graph):
        """Test inserting a simple flat object."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "alice", "age": "30"}')
        
        result = g.v("alice").inc().out("name").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "alice"

    def test_nested_object(self, clean_graph):
        """Test inserting an object with nested objects."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "bob", "address": {"city": "toronto", "country": "canada"}}')
        
        # Can traverse to nested property
        result = g.v("bob").inc().out("address").out("city").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "toronto"

    def test_array_of_primitives(self, clean_graph):
        """Test inserting an object with an array of primitive values."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "charlie", "colors": ["red", "green", "blue"]}')
        
        # colors points to a list node, then list items have the color values
        result = g.v("charlie").inc().out("colors").out("colors").all()
        assert len(result["result"]) == 3
        color_ids = {r["id"] for r in result["result"]}
        assert color_ids == {"red", "green", "blue"}

    def test_array_of_objects(self, clean_graph):
        """Test inserting an object with an array of objects."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "dave", "friends": [{"name": "eve"}, {"name": "frank"}]}')
        
        result = g.v("dave").inc().out("friends").out().out("name").all()
        assert len(result["result"]) == 2
        names = {r["id"] for r in result["result"]}
        assert names == {"eve", "frank"}

    def test_deeply_nested_structure(self, clean_graph):
        """Test inserting deeply nested JSON."""
        g = clean_graph
        g.putj('''{
            "_id": "1", 
            "name": "root",
            "level1": {
                "level2": {
                    "level3": {
                        "value": "deep"
                    }
                }
            }
        }''')
        
        result = g.v("root").inc().out("level1").out("level2").out("level3").out("value").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "deep"


class TestUpdateJson:
    """Tests for updatej - updating existing JSON objects in the graph."""

    def test_update_simple_property(self, clean_graph):
        """Test updating a simple property replaces the old value."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "alice", "city": "toronto"}')
        
        # Verify initial state
        result = g.v("alice").inc().out("city").all()
        assert result["result"][0]["id"] == "toronto"
        
        # Update
        g.updatej('{"_id": "1", "name": "alice", "city": "vancouver"}')
        
        # Verify update - should have only one city now
        result = g.v("alice").inc().out("city").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "vancouver"

    def test_update_nested_object(self, clean_graph):
        """Test updating a nested object replaces properties correctly."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "bob", "address": {"city": "toronto", "country": "canada"}}')
        
        # Update the address
        g.updatej('{"_id": "1", "name": "bob", "address": {"city": "montreal", "country": "canada"}}')
        
        result = g.v("bob").inc().out("address").out("city").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "montreal"

    def test_update_array_replaces_list_node(self, clean_graph):
        """Test updating an array creates a new list node with new items."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "charlie", "tags": ["a", "b"]}')
        
        # Update with new tags
        g.updatej('{"_id": "1", "name": "charlie", "tags": ["x", "y", "z"]}')
        
        # tags points to a list node, then list items have the tag values
        result = g.v("charlie").inc().out("tags").out("tags").all()
        assert len(result["result"]) == 3
        tags = {r["id"] for r in result["result"]}
        assert tags == {"x", "y", "z"}

    def test_update_by_nested_id(self, clean_graph):
        """Test updating a nested object directly by its _id."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "dave", "location": {"_id": "loc1", "city": "ottawa"}}')
        
        # Verify initial
        result = g.v("dave").inc().out("location").out("city").all()
        assert result["result"][0]["id"] == "ottawa"
        
        # Update the location directly by its _id
        g.updatej('{"_id": "loc1", "city": "quebec"}')
        
        # Should be updated
        result = g.v("dave").inc().out("location").out("city").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "quebec"

    def test_update_preserves_unmentioned_edges(self, clean_graph):
        """Test that updating one property doesn't affect unrelated edges from same node."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "eve", "city": "london", "country": "uk"}')
        
        # Update only city
        g.updatej('{"_id": "1", "city": "manchester"}')
        
        # City should be updated
        result = g.v().has("city", "manchester").all()
        assert len(result["result"]) == 1
        
        # Note: country edge still exists on the node (updatej doesn't delete unmentioned properties)
        # This is expected behavior - updatej updates what you specify

    def test_update_object_in_array_without_id(self, clean_graph):
        """Test updating when array items don't have _id - creates new items."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "frank", "items": [{"value": "old1"}, {"value": "old2"}]}')
        
        # Update with new items (no _id means new blank nodes)
        g.updatej('{"_id": "1", "name": "frank", "items": [{"value": "new1"}, {"value": "new2"}, {"value": "new3"}]}')
        
        result = g.v("frank").inc().out("items").out().out("value").all()
        assert len(result["result"]) == 3
        values = {r["id"] for r in result["result"]}
        assert values == {"new1", "new2", "new3"}


class TestJsonEdgeCases:
    """Edge cases and potential problem scenarios."""

    def test_empty_object(self, clean_graph):
        """Test inserting an object with only _id."""
        g = clean_graph
        g.putj('{"_id": "empty"}')
        
        # Should be able to find the node
        result = g.v("_:_id_empty").all()
        assert len(result["result"]) == 1

    def test_empty_array(self, clean_graph):
        """Test inserting an object with an empty array."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "test", "items": []}')
        
        # Name should exist
        result = g.v("test").inc().out("name").all()
        assert len(result["result"]) == 1

    def test_multiple_objects_same_structure(self, clean_graph):
        """Test inserting multiple objects with same structure."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "alice", "role": "admin"}')
        g.putj('{"_id": "2", "name": "bob", "role": "user"}')
        g.putj('{"_id": "3", "name": "charlie", "role": "admin"}')
        
        # Find all admins
        result = g.v("admin").inc().out("name").all()
        assert len(result["result"]) == 2
        names = {r["id"] for r in result["result"]}
        assert names == {"alice", "charlie"}

    def test_update_then_query_consistency(self, clean_graph):
        """Test that multiple updates maintain graph consistency."""
        g = clean_graph
        
        # Initial insert
        g.putj('{"_id": "1", "name": "user", "status": "active"}')
        
        # Multiple updates
        g.updatej('{"_id": "1", "status": "pending"}')
        g.updatej('{"_id": "1", "status": "inactive"}')
        g.updatej('{"_id": "1", "status": "active"}')
        
        # Should have only the final status
        result = g.v("user").inc().out("status").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "active"

    def test_shared_values_between_objects(self, clean_graph):
        """Test that shared values between objects work correctly."""
        g = clean_graph
        g.putj('{"_id": "1", "name": "alice", "team": "engineering"}')
        g.putj('{"_id": "2", "name": "bob", "team": "engineering"}')
        
        # Both should point to same team value
        result = g.v("engineering").inc().out("name").all()
        assert len(result["result"]) == 2
        names = {r["id"] for r in result["result"]}
        assert names == {"alice", "bob"}

    def test_update_does_not_affect_other_objects(self, clean_graph):
        """Test that updating one object doesn't corrupt another."""
        g = clean_graph
        
        # Two objects sharing a common value
        g.putj('{"_id": "1", "name": "alice", "city": "toronto"}')
        g.putj('{"_id": "2", "name": "bob", "city": "toronto"}')
        
        # Update alice's city
        g.updatej('{"_id": "1", "name": "alice", "city": "vancouver"}')
        
        # Bob should still be in toronto
        result = g.v("bob").inc().out("city").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "toronto"
        
        # Alice should be in vancouver
        result = g.v("alice").inc().out("city").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "vancouver"

    def test_complex_nested_update(self, clean_graph):
        """Test complex nested structure update."""
        g = clean_graph
        
        g.putj('''{
            "_id": "company1",
            "name": "TechCorp",
            "departments": [
                {"name": "Engineering", "head": "alice"},
                {"name": "Marketing", "head": "bob"}
            ]
        }''')
        
        # Update with new departments
        g.updatej('''{
            "_id": "company1",
            "name": "TechCorp",
            "departments": [
                {"name": "Engineering", "head": "charlie"},
                {"name": "Sales", "head": "dave"}
            ]
        }''')
        
        # Check new department heads
        result = g.v("TechCorp").inc().out("departments").out().out("head").all()
        assert len(result["result"]) == 2
        heads = {r["id"] for r in result["result"]}
        assert heads == {"charlie", "dave"}


class TestJsonWithIds:
    """Tests specifically for _id handling."""

    def test_id_creates_predictable_node(self, clean_graph):
        """Test that _id creates a predictable node identifier."""
        g = clean_graph
        g.putj('{"_id": "user123", "name": "test"}')
        
        # Should be able to query by the blank node id
        result = g.v("_:_id_user123").out("name").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "test"

    def test_update_finds_by_id(self, clean_graph):
        """Test that updatej correctly finds and updates by _id."""
        g = clean_graph
        g.putj('{"_id": "x", "value": "original"}')
        g.updatej('{"_id": "x", "value": "updated"}')
        
        result = g.v("_:_id_x").out("value").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "updated"

    def test_nested_object_with_id_during_put(self, clean_graph):
        """Test inserting nested objects that have their own _id."""
        g = clean_graph
        g.putj('{"_id": "parent", "child": {"_id": "child1", "name": "nested"}}')
        
        # Can access child by its own _id
        result = g.v("_:_id_child1").out("name").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "nested"
        
        # Can also traverse from parent
        result = g.v("_:_id_parent").out("child").out("name").all()
        assert len(result["result"]) == 1
        assert result["result"][0]["id"] == "nested"


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""
Tests for:
- Vertex.get_dict() and Vertex.__str__()
- Graph with custom cog_path_prefix
- Graph.load_csv with id_column_name validation
- Graph.__adjacent_vertices with direction='in'
- Graph.hasr() method (reverse has)
- Graph.getv() to retrieve views
- View.render() and View.__str__()
"""
from cog.torque import Graph, Vertex, View, BlankNode
from cog import config
import unittest
import os
import shutil
import json


DIR_NAME = "TorqueCoverageTest"


class TestTorqueCoverage(unittest.TestCase):
    """Tests to increase coverage of torque.py"""

    @classmethod
    def setUpClass(cls):
        cls.db_path = "/tmp/" + DIR_NAME
        cls._cleanup_path(cls.db_path)
        os.makedirs(cls.db_path)

    @classmethod
    def tearDownClass(cls):
        cls._cleanup_path(cls.db_path)
        print("*** deleted torque coverage test data.")

    @staticmethod
    def _cleanup_path(path):
        if os.path.exists(path):
            shutil.rmtree(path)

    def test_vertex_get_dict(self):
        """Test Vertex.get_dict() returns the __dict__"""
        vertex = Vertex("test_id")
        vertex.tags["tag1"] = "value1"
        vertex.set_edge("follows")
        
        d = vertex.get_dict()
        
        self.assertIsInstance(d, dict)
        self.assertEqual(d["id"], "test_id")
        self.assertIn("tag1", d["tags"])
        self.assertEqual(d["tags"]["tag1"], "value1")
        self.assertIn("follows", d["edges"])

    def test_vertex_str(self):
        """Test Vertex.__str__() - verifies set serialization raises TypeError.
        
        Known limitation: Vertex.edges is a set which isn't directly JSON serializable.
        This test documents the current behavior where __str__ raises TypeError
        when edges contains values.
        """
        vertex = Vertex("str_test")
        vertex.tags["name"] = "alice"
        vertex.set_edge("follows")  # Add an edge to make edges non-empty
        
        # Vertex.__str__ uses json.dumps which cannot serialize sets
        with self.assertRaises(TypeError):
            str(vertex)

    def test_graph_with_custom_path_prefix(self):
        """Test Graph creation with custom cog_path_prefix"""
        custom_path = "/tmp"
        g = Graph("custom_prefix_graph", cog_home=DIR_NAME, cog_path_prefix=custom_path)
        
        # Should be able to put and get data
        g.put("alice", "follows", "bob")
        result = g.v("alice").out("follows").all()
        
        self.assertEqual(result["result"][0]["id"], "bob")
        g.close()

    def test_load_csv_requires_id_column(self):
        """Test load_csv raises exception when id_column_name is None"""
        g = Graph("csv_test_graph", cog_home=DIR_NAME)
        
        csv_path = os.path.join(os.path.dirname(__file__), 'test-data', 'books.csv')
        
        with self.assertRaises(Exception) as context:
            g.load_csv(csv_path, None)
        
        self.assertIn("id_column_name must not be None", str(context.exception))
        g.close()

    def test_hasr_method(self):
        """Test hasr() - reverse has traversal"""
        g = Graph("hasr_test_graph", cog_home=DIR_NAME)
        
        g.put("alice", "follows", "bob")
        g.put("charlie", "follows", "bob")
        g.put("dani", "follows", "bob")
        
        # hasr should find vertices that have incoming edges from a specific vertex
        # Start from bob, follow "follows" predicate backwards
        result = g.v("bob").hasr("follows", "alice").all()
        
        # bob should be returned because alice follows bob
        self.assertEqual(len(result["result"]), 1)
        self.assertEqual(result["result"][0]["id"], "bob")
        
        g.close()

    def test_adjacent_vertices_in_direction(self):
        """Test __adjacent_vertices with 'in' direction via inc()"""
        g = Graph("adj_in_test_graph", cog_home=DIR_NAME)
        
        g.put("alice", "follows", "bob")
        g.put("charlie", "follows", "bob")
        
        # inc() uses direction='in' internally
        result = g.v("bob").inc("follows").all()
        
        ids = {r["id"] for r in result["result"]}
        self.assertIn("alice", ids)
        self.assertIn("charlie", ids)
        
        g.close()

    def test_view_creation_and_getv(self):
        """Test view creation and getv() to retrieve it"""
        g = Graph("view_test_graph", cog_home=DIR_NAME)
        
        g.put("alice", "follows", "bob")
        g.put("bob", "follows", "charlie")
        
        # Create a view
        view = g.v().tag("from").out("follows").tag("to").view("test_view")
        
        self.assertIsNotNone(view)
        self.assertIsInstance(view, View)
        
        # Use getv to retrieve the view
        retrieved_view = g.getv("test_view")
        
        self.assertIsNotNone(retrieved_view)
        self.assertIsInstance(retrieved_view, View)
        self.assertIn("test_view", retrieved_view.url)
        
        g.close()

    def test_view_str_method(self):
        """Test View.__str__() returns the url"""
        g = Graph("view_str_test_graph", cog_home=DIR_NAME)
        
        g.put("x", "rel", "y")
        view = g.v().tag("from").out("rel").tag("to").view("str_view")
        
        str_repr = str(view)
        
        self.assertIn("str_view", str_repr)
        self.assertTrue(str_repr.endswith(".html"))
        
        g.close()

    def test_lsv_lists_views(self):
        """Test lsv() lists all views"""
        g = Graph("lsv_test_graph", cog_home=DIR_NAME)
        
        g.put("a", "b", "c")
        g.v().tag("from").out("b").tag("to").view("view1")
        g.v().tag("from").out("b").tag("to").view("view2")
        
        views = g.lsv()
        
        self.assertIn("view1", views)
        self.assertIn("view2", views)
        
        g.close()

    def test_getv_nonexistent_raises(self):
        """Test getv() raises assertion for non-existent view"""
        g = Graph("getv_fail_test", cog_home=DIR_NAME)
        
        with self.assertRaises(AssertionError):
            g.getv("nonexistent_view")
        
        g.close()

    def test_blank_node_is_id(self):
        """Test BlankNode.is_id() class method"""
        bn = BlankNode("test123")
        
        self.assertTrue(BlankNode.is_id(str(bn)))
        self.assertFalse(BlankNode.is_id("regular_id"))
        self.assertFalse(BlankNode.is_id("_:other_format"))

    def test_blank_node_str(self):
        """Test BlankNode.__str__()"""
        bn = BlankNode("myid")
        
        str_repr = str(bn)
        
        self.assertTrue(str_repr.startswith("_:"))
        self.assertIn("myid", str_repr)

    def test_inc_with_predicates_list(self):
        """Test inc() with a list of predicates"""
        g = Graph("inc_list_test", cog_home=DIR_NAME)
        
        g.put("alice", "follows", "bob")
        g.put("charlie", "likes", "bob")
        
        result = g.v("bob").inc(["follows", "likes"]).all()
        
        ids = {r["id"] for r in result["result"]}
        self.assertIn("alice", ids)
        self.assertIn("charlie", ids)
        
        g.close()

    def test_view_persist(self):
        """Test View.persist() saves to file"""
        g = Graph("view_persist_test", cog_home=DIR_NAME)
        
        g.put("p", "q", "r")
        view = g.v().tag("from").out("q").tag("to").view("persist_view")
        
        # Check the file exists
        self.assertTrue(os.path.isfile(view.url))
        
        # Read the file content
        with open(view.url, 'r') as f:
            content = f.read()
        
        self.assertIn("vis.DataSet", content)  # vis.js library usage
        
        g.close()


if __name__ == '__main__':
    unittest.main()



from cog.torque import Graph, ASC, DESC
import unittest
import os
import shutil

DIR_NAME = "TorqueTest"


# target api set
# in("predicate"), out("predicate"), all, count, tag
# next release .forEach(function), this can do filter, adding values etc.

# // The working set of this is bob and dani
# g.V("<charlie>").Out("<follows>").All()
# // The working set of this is fred, as alice follows bob and bob follows fred.
# g.V("<alice>").Out("<follows>").Out("<follows>").All()
# // Finds all things dani points at. Result is bob, greg and cool_person
# g.V("<dani>").Out().All()
# // Finds all things dani points at on the status linkage.
# // Result is bob, greg and cool_person
# g.V("<dani>").Out(["<follows>", "<status>"]).All()
# // Finds all things dani points at on the status linkage, given from a separate query path.
# // Result is {"id": "cool_person", "pred": "<status>"}
# g.V("<dani>").Out(g.V("<status>"), "pred").All()

# https://docs.janusgraph.org/latest/gremlin.html
# https://tinkerpop.apache.org/gremlin.html


def ordered(obj):
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in list(obj.items()))
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    else:
        return obj


class TorqueTest(unittest.TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):

        if os.path.exists("/tmp/" + DIR_NAME):
            shutil.rmtree("/tmp/" + DIR_NAME)

        if not os.path.exists("/tmp/" + DIR_NAME):
            os.mkdir("/tmp/" + DIR_NAME)

        data_dir = "test/test-data/test.nq"
        # choose appropriate path based on where the test is called from.
        if os.path.exists("test-data/test.nq"):
            data_dir = "test-data/test.nq"

        TorqueTest.g = Graph(graph_name="people", cog_home=DIR_NAME)
        TorqueTest.g.load_triples(data_dir, "people")
        # print TorqueTest.g.v().all()
        print(">>> test setup complete.\n")

    def test_torque_1(self):
        self.assertEqual(1, TorqueTest.g.v("<alice>").out().count())

    def test_torque_2(self):
        expected = {'result': [{'source': 'cool_person', 'id': 'cool_person'}, {'source': '<fred>', 'id': '<fred>'}]}
        actual = TorqueTest.g.v("<bob>").out().tag("source").all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_torque_3(self):
        expected = {"result": [{"source": "<fred>", "id": "<greg>", "target": "<greg>"}]}
        actual = TorqueTest.g.v("<bob>").out().tag("source").out().tag("target").all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_torque_4(self):
        expected = {'result': [{'source': 'cool_person', 'id': '<bob>', 'target': '<bob>'},
                               {'source': 'cool_person', 'id': '<dani>', 'target': '<dani>'},
                               {'source': 'cool_person', 'id': '<greg>', 'target': '<greg>'},
                               {'source': '<fred>', 'id': '<bob>', 'target': '<bob>'},
                               {'source': '<fred>', 'id': '<emily>', 'target': '<emily>'}]}
        actual = TorqueTest.g.v("<bob>").out().tag("source").inc().tag("target").all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_torque_5(self):
        expected = {'result': [{'source': '<greg>', 'id': '<dani>', 'target': '<dani>'},
                               {'source': '<greg>', 'id': '<fred>', 'target': '<fred>'}]}
        actual = TorqueTest.g.v("<fred>").out().tag("source").inc().tag("target").all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_torque_6(self):
        expected = {'result': [{'source': '<greg>', 'id': '<dani>', 'target': '<dani>'},
                               {'source': '<greg>', 'id': '<fred>', 'target': '<fred>'}]}
        actual = TorqueTest.g.v("<fred>").out().tag("source").inc().tag("target").all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_torque_7(self):
        expected = {'result': [{'source': '<greg>', 'id': 'cool_person', 'target': 'cool_person'}]}
        actual = TorqueTest.g.v("<fred>").out().tag("source").out().tag("target").all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_torque_8(self):
        expected = {'result': [{'source': '<greg>', 'id': 'cool_person', 'target': 'cool_person'},
                               {'source': '<greg>', 'id': '<bob>', 'target': '<bob>'},
                               {'source': '<greg>', 'id': '<greg>', 'target': '<greg>'},
                               {'source': '<greg>', 'id': '<greg>', 'target': '<greg>'}]}
        actual = TorqueTest.g.v("<fred>").out().tag("source").inc().out().tag("target").all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_torque_9(self):
        expected = {'result': [{'source': '<fred>', 'id': '<fred>'}]}
        actual = TorqueTest.g.v("<bob>").out(["<follows>"]).tag("source").all()
        self.assertTrue(ordered(expected) == ordered(actual))

    def test_torque_10(self):
        expected = {'result': [{'source': '<fred>', 'id': '<fred>'}, {'source': 'cool_person', 'id': 'cool_person'}]}
        actual = TorqueTest.g.v("<bob>").out(["<follows>", "<status>"]).tag("source").all()
        self.assertEqual(ordered(expected), ordered(actual))

    # bad predicates, should not break test
    def test_torque_11(self):
        expected = {'result': []}
        actual = TorqueTest.g.v("<bob>").out(["<follows>zzz", "<status>zzz"]).tag("source").all()
        self.assertTrue(ordered(expected) == ordered(actual))

    def test_torque_12(self):
        actual = TorqueTest.g.scan(3, 'v')
        print(actual)
        self.assertTrue(3 == len(actual['result']))
        with self.assertRaises(AssertionError):
            TorqueTest.g.scan('v', 3)

    def test_torque_13(self):
        expected = {'result': [{'id': '<greg>'}, {'id': '<dani>'}, {'id': '<bob>'}]}
        actual = TorqueTest.g.v().inc(["<status>"]).all()
        self.assertTrue(expected == actual)

    def test_torque_14(self):
        expected = {
            'result': [{'id': '<fred>'}, {'id': '<dani>'}, {'id': '<charlie>'}, {'id': '<dani>'}, {'id': '<charlie>'},
                       {'id': '<alice>'}]}
        actual = TorqueTest.g.v().inc(["<status>"]).inc("<follows>").all()
        self.assertTrue(expected == actual)

    def test_torque_15(self):
        view = TorqueTest.g.v("<bob>").out().tag("from").inc().tag("to").view("bob_view")
        print(view.url)
        self.assertTrue(view.url.endswith("bob_view.html"))
        self.assertEqual(['bob_view'], TorqueTest.g.lsv())
        view = TorqueTest.g.v("<dani>").tag("from").out().tag("to").view("dani_view")
        self.assertEqual(['bob_view', 'dani_view'], sorted(TorqueTest.g.lsv()))

    def test_torque_16(self):
        expected = {'result': [{'id': '<bob>'}]}
        actual = TorqueTest.g.v("<charlie>").out("<follows>").has("<follows>", "<fred>").all()
        self.assertTrue(expected == actual)

    def test_torque_17(self):
        expected = {'result': [{'id': '<dani>'}, {'id': '<alice>'}, {'id': '<charlie>'}]}
        actual = TorqueTest.g.v().has("<follows>", "<bob>").all()
        self.assertTrue(expected == actual)

    def test_torque_18(self):
        expected = {'result': [{'id': '<bob>'}, {'id': '<dani>'}, {'id': '<greg>'}]}
        actual = TorqueTest.g.v().has("<status>", 'cool_person').all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_torque_19(self):
        expected = {'result': [{'id': '<fred>', 'edges': ['<follows>']}, {'id': 'cool_person', 'edges': ['<status>']}]}
        actual = TorqueTest.g.v("<bob>").out().all('e')
        self.assertEqual(ordered(expected), ordered(actual))

    def test_torque_20(self):
        expected = {'result': [{'id': '<bob>'}, {'id': '<emily>'}]}
        actual = TorqueTest.g.v().has("<follows>", "<fred>").all('e')
        self.assertTrue(expected == actual)

    def test_torque_21(self):
        expected = {'result': [{'id': '<dani>', 'edges': ['<follows>']}, {'id': '<charlie>', 'edges': ['<follows>']},
                               {'id': '<alice>', 'edges': ['<follows>']}]}
        actual = TorqueTest.g.v().has("<follows>", "<fred>").inc().all('e')
        self.assertTrue(expected == actual)

    def test_torque_22(self):
        expected = {'result': [{'id': '<bob>'}]}
        actual = TorqueTest.g.v().hasr("<follows>", "<alice>").all()
        self.assertEqual(expected, actual)

    # ============================================================
    # Multi-tag tests
    # ============================================================

    def test_multi_tag_with_list(self):
        """Test that tag() accepts a list of tag names and tags all of them."""
        expected = {'result': [{'source': 'cool_person', 'target': 'cool_person', 'id': 'cool_person'},
                               {'source': '<fred>', 'target': '<fred>', 'id': '<fred>'}]}
        actual = TorqueTest.g.v("<bob>").out().tag(["source", "target"]).all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_multi_tag_single_string_still_works(self):
        """Test that tag() still works with a single string (backward compatibility)."""
        expected = {'result': [{'source': 'cool_person', 'id': 'cool_person'},
                               {'source': '<fred>', 'id': '<fred>'}]}
        actual = TorqueTest.g.v("<bob>").out().tag("source").all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_multi_tag_empty_list(self):
        """Test that tag() with empty list does not add any tags."""
        expected = {'result': [{'id': 'cool_person'}, {'id': '<fred>'}]}
        actual = TorqueTest.g.v("<bob>").out().tag([]).all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_multi_tag_chained(self):
        """Test chaining: single tag followed by multi-tag."""
        expected = {'result': [{'first': '<fred>', 'second': '<greg>', 'third': '<greg>', 'id': '<greg>'}]}
        actual = TorqueTest.g.v("<bob>").out("<follows>").tag("first").out().tag(["second", "third"]).all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_multi_tag_with_traversal(self):
        """Test multi-tag combined with out() accepting a list of predicates."""
        expected = {'result': [{'both': 'cool_person', 'id': 'cool_person'},
                               {'both': '<fred>', 'id': '<fred>'}]}
        actual = TorqueTest.g.v("<bob>").out(["<follows>", "<status>"]).tag(["both"]).all()
        self.assertEqual(ordered(expected), ordered(actual))

    def test_multi_tag_preserves_previous_tags(self):
        """Test that multi-tagging preserves tags from earlier in the traversal."""
        expected = {'result': [{'start': '<bob>', 'end1': '<greg>', 'end2': '<greg>', 'id': '<greg>'}]}
        actual = TorqueTest.g.v("<bob>").tag("start").out("<follows>").out().tag(["end1", "end2"]).all()
        self.assertEqual(ordered(expected), ordered(actual))

    # ============================================================
    # Order tests
    # ============================================================

    def test_order_default_ascending(self):
        """Test that order() without arguments sorts ascending by default."""
        actual = TorqueTest.g.v("<bob>").out().order().all()
        ids = [r['id'] for r in actual['result']]
        self.assertEqual(ids, sorted(ids))

    def test_order_ascending(self):
        """Test that order(ASC) sorts ascending."""
        actual = TorqueTest.g.v("<bob>").out().order(ASC).all()
        ids = [r['id'] for r in actual['result']]
        self.assertEqual(ids, sorted(ids))

    def test_order_descending(self):
        """Test that order(DESC) sorts descending."""
        actual = TorqueTest.g.v("<bob>").out().order(DESC).all()
        ids = [r['id'] for r in actual['result']]
        self.assertEqual(ids, sorted(ids, reverse=True))

    def test_order_with_string_literals(self):
        """Test that order works with string literals 'asc' and 'desc'."""
        actual_asc = TorqueTest.g.v("<bob>").out().order("asc").all()
        actual_desc = TorqueTest.g.v("<bob>").out().order("desc").all()
        ids_asc = [r['id'] for r in actual_asc['result']]
        ids_desc = [r['id'] for r in actual_desc['result']]
        self.assertEqual(ids_asc, sorted(ids_asc))
        self.assertEqual(ids_desc, sorted(ids_desc, reverse=True))

    def test_order_placed_before_tag(self):
        """Test that order() can be placed before tag() in the query chain."""
        actual = TorqueTest.g.v("<bob>").out().order().tag("source").all()
        ids = [r['id'] for r in actual['result']]
        self.assertEqual(ids, sorted(ids))
        # Verify tags are still applied
        for r in actual['result']:
            self.assertIn('source', r)

    def test_order_on_empty_result(self):
        """Test that order() handles empty result sets gracefully."""
        actual = TorqueTest.g.v("<bob>").out(["<nonexistent>"]).order().all()
        self.assertEqual(actual, {'result': []})

    @classmethod
    def tearDownClass(cls):
        TorqueTest.g.close()
        shutil.rmtree("/tmp/" + DIR_NAME)
        print("*** deleted test data.")


if __name__ == '__main__':
    unittest.main()



"""
CogDB HTTP Server
=================
Enables serving Graph instances over HTTP for remote querying.
Supports multiple graphs on the same port with path-based routing.

Uses stdlib http.server for zero dependencies.
"""

from http.server import HTTPServer, BaseHTTPRequestHandler
from socketserver import ThreadingMixIn
import json
import threading
import time
import re
import socket
from urllib.parse import urlparse

from cog.templates import render_index_page, render_graph_row, render_status_page


# Version pulled dynamically when available
try:
    from importlib.metadata import version as pkg_version
    COGDB_VERSION = pkg_version('cogdb')
except Exception:
    COGDB_VERSION = "dev"


# Global registry of servers by port
_server_registry = {}  # port -> CogDBServer
_registry_lock = threading.Lock()


class ThreadingHTTPServer(ThreadingMixIn, HTTPServer):
    """Thread-per-request HTTP server."""
    daemon_threads = True
    allow_reuse_address = True


class CogDBRequestHandler(BaseHTTPRequestHandler):
    """
    HTTP request handler for CogDB server.
    Supports path-based routing for multiple graphs.
    """
    
    # Suppress default logging
    def log_message(self, format, *args):
        pass
    
    def _send_json(self, data, status=200):
        """Send a JSON response."""
        body = json.dumps(data).encode('utf-8')
        self.send_response(status)
        self.send_header('Content-Type', 'application/json')
        self.send_header('Content-Length', len(body))
        self.send_header('Access-Control-Allow-Origin', '*')
        self.end_headers()
        self.wfile.write(body)
    
    def _send_html(self, html, status=200):
        """Send an HTML response."""
        body = html.encode('utf-8')
        self.send_response(status)
        self.send_header('Content-Type', 'text/html; charset=utf-8')
        self.send_header('Content-Length', len(body))
        self.end_headers()
        self.wfile.write(body)
    
    def _read_json_body(self):
        """Read and parse JSON request body."""
        content_length = int(self.headers.get('Content-Length', 0))
        if content_length == 0:
            return None
        body = self.rfile.read(content_length)
        return json.loads(body.decode('utf-8'))
    
    def _get_local_ip(self):
        """Get the local IP address of this machine."""
        try:
            # Create a socket to determine local IP
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(('8.8.8.8', 80))
            ip = s.getsockname()[0]
            s.close()
            return ip
        except Exception:
            return '127.0.0.1'
    
    def _parse_path(self):
        """Parse the path to extract graph name and action."""
        path = self.path.rstrip('/')
        if path == '' or path == '/':
            return None, 'index'
        
        parts = path.split('/')
        # /graph_name or /graph_name/action
        if len(parts) >= 2:
            graph_name = parts[1]
            action = parts[2] if len(parts) >= 3 else 'status'
            return graph_name, action
        
        return None, 'index'
    
    def _get_share_url(self):
        """Get the share URL from X-Share-Url header (set by relay for remote access)."""
        return self.headers.get('X-Share-Url', '')
    
    def _get_graph_state(self, graph_name):
        """Get the state for a specific graph."""
        graphs = self.server.cog_graphs
        if graph_name not in graphs:
            return None
        return graphs[graph_name]
    
    def do_OPTIONS(self):
        """Handle CORS preflight."""
        self.send_response(200)
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')
        self.send_header('Access-Control-Allow-Headers', 'Content-Type')
        self.end_headers()
    
    def do_GET(self):
        """Handle GET requests."""
        graph_name, action = self._parse_path()
        
        if action == 'index':
            self._handle_index_page()
        elif graph_name:
            state = self._get_graph_state(graph_name)
            if not state:
                self._send_json({'ok': False, 'error': f'Graph not found: {graph_name}'}, 404)
                return
            
            if action == 'status':
                self._handle_status_page(graph_name, state)
            elif action == 'stats':
                self._handle_stats(graph_name, state)
            else:
                self._send_json({'ok': False, 'error': 'Not found'}, 404)
        else:
            self._send_json({'ok': False, 'error': 'Not found'}, 404)
    
    def do_POST(self):
        """Handle POST requests."""
        graph_name, action = self._parse_path()
        
        if not graph_name:
            self._send_json({'ok': False, 'error': 'Graph name required in path'}, 400)
            return
        
        state = self._get_graph_state(graph_name)
        if not state:
            self._send_json({'ok': False, 'error': f'Graph not found: {graph_name}'}, 404)
            return
        
        if action == 'query':
            self._handle_query(graph_name, state)
        elif action == 'mutate':
            self._handle_mutate(graph_name, state)
        else:
            self._send_json({'ok': False, 'error': 'Not found'}, 404)
    
    def _handle_index_page(self):
        """Render index page listing all available graphs."""
        graphs = self.server.cog_graphs
        start_time = self.server.cog_start_time
        
        # Calculate uptime
        uptime_secs = int(time.time() - start_time)
        hours, remainder = divmod(uptime_secs, 3600)
        minutes, seconds = divmod(remainder, 60)
        uptime_str = f"{hours}h {minutes}m {seconds}s"
        
        # Build graph list HTML
        graph_rows = []
        for name, state in graphs.items():
            try:
                node_count = state['graph'].v().count()
            except Exception:
                node_count = 0
            mode = "rw" if state['writable'] else "ro"
            graph_rows.append(render_graph_row(name, node_count, mode))
        
        graphs_html = ''.join(graph_rows) if graph_rows else '<div style="color: #666;">No graphs registered</div>'
        
        html = render_index_page(
            version=COGDB_VERSION,
            local_ip=self._get_local_ip(),
            port=self.server.server_address[1],
            graphs_html=graphs_html,
            uptime_str=uptime_str,
            share_url=self._get_share_url()
        )
        self._send_html(html)
    
    def _handle_status_page(self, graph_name, state):
        """Render terminal-style status page for a specific graph."""
        graph = state['graph']
        start_time = state['start_time']
        queries_served = state['queries_served']
        last_query_time = state['last_query_time']
        writable = state['writable']
        
        # Calculate uptime
        uptime_secs = int(time.time() - start_time)
        hours, remainder = divmod(uptime_secs, 3600)
        minutes, seconds = divmod(remainder, 60)
        uptime_str = f"{hours}h {minutes}m {seconds}s"
        
        # Get node/edge counts
        try:
            node_count = graph.v().count()
            edge_count = len(graph.all_predicates) if hasattr(graph, 'all_predicates') else 0
        except Exception:
            node_count = 0
            edge_count = 0
        
        # Last query ago
        if last_query_time:
            ago = int(time.time() - last_query_time)
            last_query_str = f"{ago}s ago"
        else:
            last_query_str = "never"
        
        mode_str = "writable" if writable else "read-only"
        
        # Get instance ID
        try:
            instance_id = graph.cog.instance_id if hasattr(graph, 'cog') else 'N/A'
        except Exception:
            instance_id = 'N/A'
        
        html = render_status_page(
            version=COGDB_VERSION,
            local_ip=self._get_local_ip(),
            port=self.server.server_address[1],
            graph_name=graph_name,
            instance_id=instance_id,
            node_count=node_count,
            edge_count=edge_count,
            uptime_str=uptime_str,
            queries_served=queries_served,
            last_query_str=last_query_str,
            mode_str=mode_str,
            share_url=self._get_share_url()
        )
        self._send_html(html)
    
    def _handle_stats(self, graph_name, state):
        """Return JSON statistics for a graph."""
        graph = state['graph']
        start_time = state['start_time']
        
        try:
            node_count = graph.v().count()
        except Exception:
            node_count = 0
        
        edge_count = len(graph.all_predicates) if hasattr(graph, 'all_predicates') else 0
        
        stats = {
            'version': COGDB_VERSION,
            'graph_name': graph_name,
            'nodes': node_count,
            'edges': edge_count,
            'uptime_seconds': int(time.time() - start_time),
            'queries_served': state['queries_served'],
            'writable': state['writable']
        }
        self._send_json(stats)
    
    def _handle_query(self, graph_name, state):
        """Execute a Torque query on a specific graph."""
        try:
            body = self._read_json_body()
            if not body or 'q' not in body:
                self._send_json({'ok': False, 'error': 'Missing query parameter "q"'}, 400)
                return
            
            query_str = body['q']
            graph = state['graph']
            query_lock = state.get('query_lock')
            
            # Execute the query safely with lock for thread safety
            if query_lock:
                with query_lock:
                    result = self._execute_query(graph, query_str)
            else:
                result = self._execute_query(graph, query_str)
            
            # Update stats
            state['queries_served'] += 1
            state['last_query_time'] = time.time()
            
            self._send_json({'ok': True, 'result': result.get('result', result)})
            
        except Exception as e:
            self._send_json({'ok': False, 'error': str(e)}, 400)
    
    def _execute_query(self, graph, query_str):
        """Safely execute a Torque query string."""
        # Validate query structure - must start with allowed methods
        allowed_starts = ['v(', 'scan(']
        query_stripped = query_str.strip()
        
        if not any(query_stripped.startswith(s) for s in allowed_starts):
            raise ValueError(f"Query must start with one of: {allowed_starts}")
        
        # SECURITY: Block dunder attributes (prevents __globals__, __init__ RCE attacks)
        if '__' in query_str:
            raise ValueError("Query contains forbidden pattern '__'")
        
        # Validate method chain - only allow known traversal methods
        allowed_methods = {
            'v', 'out', 'inc', 'both', 'has', 'hasr', 'tag', 'back',
            'all', 'count', 'first', 'one', 'scan', 'filter', 'unique', 'limit', 'skip',
            'is_', 'bfs', 'dfs', 'sim', 'k_nearest'
        }
        
        # Extract method names from the query
        method_pattern = r'\.?([a-zA-Z_][a-zA-Z0-9_]*)\s*\('
        methods_used = set(re.findall(method_pattern, query_str))
        
        invalid_methods = methods_used - allowed_methods
        if invalid_methods:
            raise ValueError(f"Disallowed methods: {invalid_methods}")
        
        # Build and execute - use compile to check syntax first
        full_query = f"graph.{query_str}"
        try:
            compile(full_query, '<query>', 'eval')
        except SyntaxError as e:
            raise ValueError(f"Invalid query syntax: {e}")
        
        # Execute in restricted namespace
        result = eval(full_query, {"__builtins__": {}}, {"graph": graph})
        
        # Handle different return types
        if isinstance(result, dict):
            return result
        elif isinstance(result, int):
            return {'result': result}
        else:
            return {'result': []}
    
    def _handle_mutate(self, graph_name, state):
        """Handle write operations on a specific graph."""
        if not state['writable']:
            self._send_json({
                'ok': False, 
                'error': 'Write operations disabled. Start server with writable=True'
            }, 403)
            return
        
        try:
            body = self._read_json_body()
            if not body or 'op' not in body:
                self._send_json({'ok': False, 'error': 'Missing "op" parameter'}, 400)
                return
            
            op = body['op']
            args = body.get('args', [])
            graph = state['graph']
            
            if op == 'put':
                if len(args) != 3:
                    self._send_json({'ok': False, 'error': 'put requires [subject, predicate, object]'}, 400)
                    return
                graph.put(args[0], args[1], args[2])
                self._send_json({'ok': True, 'affected': 1})
                
            elif op == 'put_batch':
                if not isinstance(args, list):
                    self._send_json({'ok': False, 'error': 'put_batch requires list of [s,p,o] triples'}, 400)
                    return
                triples = [tuple(t) for t in args]
                graph.put_batch(triples)
                self._send_json({'ok': True, 'affected': len(triples)})
                
            elif op == 'delete':
                if len(args) != 3:
                    self._send_json({'ok': False, 'error': 'delete requires [subject, predicate, object]'}, 400)
                    return
                graph.delete(args[0], args[1], args[2])
                self._send_json({'ok': True, 'affected': 1})
            
            elif op == 'truncate':
                query_lock = state.get('query_lock')
                if query_lock:
                    with query_lock:
                        graph.truncate()
                else:
                    graph.truncate()
                self._send_json({'ok': True, 'message': 'Graph truncated'})
                
            else:
                self._send_json({'ok': False, 'error': f'Unknown operation: {op}'}, 400)
                
        except Exception as e:
            self._send_json({'ok': False, 'error': str(e)}, 400)


class CogDBServer:
    """
    HTTP server wrapper for CogDB Graphs.
    Supports multiple graphs on the same port.
    
    Usage:
        # Graphs register themselves via Graph.serve()
        # This class is used internally
    """
    
    def __init__(self, port=8080, host='0.0.0.0'):
        self.port = port
        self.host = host
        self.server = None
        self.thread = None
        self._running = False
        self._graphs = {}  # graph_name -> state
        self._lock = threading.Lock()
    
    def register_graph(self, graph, writable=False):
        """Register a graph to be served."""
        with self._lock:
            self._graphs[graph.graph_name] = {
                'graph': graph,
                'start_time': time.time(),
                'queries_served': 0,
                'last_query_time': None,
                'writable': writable,
                'query_lock': threading.Lock()  # Per-graph lock for thread-safe queries
            }
            # Update server's graph reference
            if self.server:
                self.server.cog_graphs = self._graphs
    
    def unregister_graph(self, graph_name):
        """Unregister a graph. Returns True if server should shutdown."""
        with self._lock:
            if graph_name in self._graphs:
                del self._graphs[graph_name]
                if self.server:
                    self.server.cog_graphs = self._graphs
            return len(self._graphs) == 0
    
    def has_graph(self, graph_name):
        """Check if a graph is registered."""
        return graph_name in self._graphs
    
    def start(self, blocking=False):
        """Start the HTTP server."""
        if self._running:
            return
        
        self.server = ThreadingHTTPServer((self.host, self.port), CogDBRequestHandler)
        self.server.cog_graphs = self._graphs
        self.server.cog_start_time = time.time()
        
        self._running = True
        
        if blocking:
            try:
                self.server.serve_forever()
            except KeyboardInterrupt:
                pass
            finally:
                self.stop()
        else:
            self.thread = threading.Thread(target=self.server.serve_forever, daemon=True)
            self.thread.start()
    
    def stop(self):
        """Stop the HTTP server and release the port."""
        if self.server:
            self.server.shutdown()
            self.server.server_close()
            self.server = None
        self._running = False
        self._graphs.clear()
    
    @property
    def url(self):
        """Get the server URL."""
        return f"http://{self.host}:{self.port}"
    
    @property
    def is_running(self):
        return self._running


def get_or_create_server(port, host='0.0.0.0'):
    """Get existing server on port or create a new one."""
    with _registry_lock:
        if port in _server_registry:
            server = _server_registry[port]
            if server.is_running:
                return server, False  # existing server
        
        # Create new server
        server = CogDBServer(port=port, host=host)
        _server_registry[port] = server
        return server, True  # new server


def unregister_from_server(port, graph_name):
    """Unregister a graph from a server. Shuts down server if empty."""
    with _registry_lock:
        if port not in _server_registry:
            return
        
        server = _server_registry[port]
        should_shutdown = server.unregister_graph(graph_name)
        
        if should_shutdown:
            server.stop()
            del _server_registry[port]


def stop_server(port=8080):
    """
    Stop the entire server on a port, unregistering all graphs.
    
    Args:
        port: Port to stop server on (default 8080)
    
    Example:
        from cog.server import stop_server
        stop_server(8080)  # Stops server and all graphs on port 8080
    """
    with _registry_lock:
        if port not in _server_registry:
            return
        
        server = _server_registry[port]
        
        # Clear _server_port on each graph so they can serve again
        for graph_name, state in list(server._graphs.items()):
            graph = state.get('graph')
            if graph and hasattr(graph, '_server_port'):
                graph._server_port = None
        
        server.stop()
        del _server_registry[port]



"""
Tests for:
- Record.get_kv_tuple()
- Index load increment on non-empty blocks
- Index put loopback and capacity handling
- Store address exceeds block size exception
- Index get loopback and EOF handling
- Store EOF handling in scanner
- Index delete key not found scenarios
- Index flush
- Store read EOF handling
- Indexer with multiple indexes
"""
from cog.core import Table, Record, Index, Store, Indexer, TableMeta, cog_hash
from cog import config
import logging
import os
from logging.config import dictConfig
import shutil
import unittest

DIR_NAME = "TestCoreCoverage"


class TestCoreCoverage(unittest.TestCase):
    """Tests to increase coverage of core.py"""

    @classmethod
    def setUpClass(cls):
        cls.db_path = "/tmp/" + DIR_NAME
        cls._cleanup_path(cls.db_path)
        os.makedirs(cls.db_path + "/test_table/")
        config.CUSTOM_COG_DB_PATH = cls.db_path
        dictConfig(config.logging_config)

    @classmethod
    def tearDownClass(cls):
        cls._cleanup_path(cls.db_path)
        print("*** deleted core coverage test data.")

    @staticmethod
    def _cleanup_path(path):
        if os.path.exists(path):
            shutil.rmtree(path)

    def test_record_get_kv_tuple(self):
        """Test Record.get_kv_tuple() returns correct tuple"""
        record = Record("my_key", "my_value")
        kv_tuple = record.get_kv_tuple()
        self.assertEqual(kv_tuple, ("my_key", "my_value"))

    def test_record_get_kv_tuple_with_list(self):
        """Test Record.get_kv_tuple() with list value"""
        record = Record("list_key", ["val1", "val2", "val3"], value_type='l')
        kv_tuple = record.get_kv_tuple()
        self.assertEqual(kv_tuple[0], "list_key")
        self.assertEqual(kv_tuple[1], ["val1", "val2", "val3"])

    def test_record_is_empty(self):
        """Test Record.is_empty() method"""
        empty_record = Record(None, None)
        self.assertTrue(empty_record.is_empty())

        non_empty_record = Record("key", "value")
        self.assertFalse(non_empty_record.is_empty())

    def test_index_flush(self):
        """Test Index.flush() method"""
        logger = logging.getLogger()
        table = Table("flush_test", "test_table", "test_flush_id", config)
        index = table.indexer.index_list[0]
        
        # Put some data
        record = Record("flush_key", "flush_value")
        position = table.store.save(record)
        index.put(record.key, position, table.store)
        
        # Flush should not raise
        index.flush()
        
        # Verify data still retrievable after flush
        retrieved = index.get("flush_key", table.store)
        self.assertIsNotNone(retrieved)
        self.assertEqual(retrieved.value, "flush_value")
        
        table.close()

    def test_index_get_nonexistent_key(self):
        """Test Index.get() returns None for non-existent key"""
        logger = logging.getLogger()
        table = Table("nonexistent_test", "test_table", "test_ne_id", config)
        index = table.indexer.index_list[0]
        
        result = index.get("definitely_not_exists", table.store)
        self.assertIsNone(result)
        
        table.close()

    def test_index_delete_nonexistent_key(self):
        """Test Index.delete() returns False for non-existent key"""
        logger = logging.getLogger()
        table = Table("delete_ne_test", "test_table", "test_del_ne_id", config)
        index = table.indexer.index_list[0]
        
        result = index.delete("key_that_does_not_exist", table.store)
        self.assertFalse(result)
        
        table.close()

    def test_index_delete_in_chain(self):
        """Test Index.delete() for key in collision chain"""
        # Save original capacity
        orig_capacity = config.INDEX_CAPACITY
        config.INDEX_CAPACITY = 4  # Force collisions
        
        logger = logging.getLogger()
        table = Table("del_chain_test", "test_table", "test_del_chain_id", config)
        store = table.store
        index = table.indexer.index_list[0]
        
        # Insert multiple records that may collide
        records = [
            Record("key_a", "value_a"),
            Record("key_b", "value_b"),
            Record("key_c", "value_c"),
            Record("key_d", "value_d"),
        ]
        
        for rec in records:
            position = store.save(rec)
            index.put(rec.key, position, store)
        
        # Delete middle key
        result = index.delete("key_b", store)
        self.assertTrue(result, "index.delete() should return True on successful deletion")
        
        # Verify deleted key is no longer accessible
        self.assertIsNone(index.get("key_b", store), "The deleted key should not be found")
        
        # Verify other keys still exist
        self.assertIsNotNone(index.get("key_a", store))
        self.assertIsNotNone(index.get("key_c", store))
        self.assertIsNotNone(index.get("key_d", store))
        
        table.close()
        config.INDEX_CAPACITY = orig_capacity

    def test_indexer_get_key_not_found_logs(self):
        """Test Indexer.get() when key not found"""
        logger = logging.getLogger()
        table = Table("indexer_nf_test", "test_table", "test_indexer_nf_id", config)
        
        result = table.indexer.get("nonexistent_key", table.store)
        self.assertIsNone(result)
        
        table.close()

    def test_store_read_eof_handling(self):
        """Test Store read behavior at boundaries"""
        logger = logging.getLogger()
        table = Table("store_eof_test", "test_table", "test_store_eof_id", config)
        store = table.store
        
        # Save a record
        record = Record("eof_key", "eof_value")
        position = store.save(record)
        
        # Read at valid position
        data = store.read(position)
        self.assertIsNotNone(data)
        
        table.close()

    def test_scanner_with_empty_index(self):
        """Test scanner on empty index"""
        logger = logging.getLogger()
        table = Table("empty_scan_test", "test_table", "test_empty_scan_id", config)
        
        count = 0
        for _ in table.indexer.scanner(table.store):
            count += 1
        
        # Empty table should yield no records
        self.assertEqual(count, 0)
        
        table.close()

    def test_scanner_with_records(self):
        """Test scanner iterates through all records"""
        logger = logging.getLogger()
        table = Table("scan_test", "test_table", "test_scan_id", config)
        store = table.store
        indexer = table.indexer
        
        # Insert records
        records = [
            Record("scan_key1", "scan_value1"),
            Record("scan_key2", "scan_value2"),
            Record("scan_key3", "scan_value3"),
        ]
        
        for rec in records:
            position = store.save(rec)
            indexer.put(rec.key, position, store)
        
        # Scan should return all records
        scanned_keys = set()
        for r in indexer.scanner(store):
            scanned_keys.add(r.key)
        
        self.assertEqual(scanned_keys, {"scan_key1", "scan_key2", "scan_key3"})
        
        table.close()

    def test_index_collision_chain_traversal(self):
        """Test get() traverses collision chain correctly"""
        orig_capacity = config.INDEX_CAPACITY
        config.INDEX_CAPACITY = 2  # Very small to force many collisions
        
        logger = logging.getLogger()
        table = Table("collision_chain_test", "test_table", "test_collision_chain_id", config)
        store = table.store
        index = table.indexer.index_list[0]
        
        # Insert many records to force collision chain
        for i in range(20):
            rec = Record(f"collision_key_{i}", f"collision_value_{i}")
            position = store.save(rec)
            index.put(rec.key, position, store)
        
        # Verify all records retrievable
        for i in range(20):
            result = index.get(f"collision_key_{i}", store)
            self.assertIsNotNone(result, f"Key collision_key_{i} should be found")
            self.assertEqual(result.value, f"collision_value_{i}")
        
        table.close()
        config.INDEX_CAPACITY = orig_capacity

    def test_indexer_delete_returns_false_for_missing_key(self):
        """Test Indexer.delete() returns False when key not found"""
        logger = logging.getLogger()
        table = Table("indexer_del_test", "test_table", "test_indexer_del_id", config)
        
        result = table.indexer.delete("missing_key", table.store)
        self.assertFalse(result)
        
        table.close()

    def test_store_batch_mode(self):
        """Test Store batch mode for deferred flushing"""
        logger = logging.getLogger()
        table = Table("batch_test", "test_table", "test_batch_id", config)
        store = table.store
        
        store.begin_batch()
        
        # Insert multiple records in batch mode
        positions = []
        for i in range(10):
            rec = Record(f"batch_key_{i}", f"batch_value_{i}")
            pos = store.save(rec)
            positions.append(pos)
        
        store.end_batch()
        
        # Verify all records readable
        for pos in positions:
            data = store.read(pos)
            self.assertIsNotNone(data)
        
        table.close()

    def test_store_sync(self):
        """Test Store.sync() forces flush"""
        logger = logging.getLogger()
        table = Table("sync_test", "test_table", "test_sync_id", config)
        store = table.store
        
        rec = Record("sync_key", "sync_value")
        store.save(rec)
        
        # sync should not raise
        store.sync()
        
        table.close()

    def test_record_str_representation(self):
        """Test Record __str__ method"""
        record = Record("str_key", "str_value", tombstone='0', store_position=100, 
                       value_type='s', key_link=50, value_link=-1)
        str_repr = str(record)
        
        self.assertIn("str_key", str_repr)
        self.assertIn("str_value", str_repr)
        self.assertIn("100", str_repr)  # store_position
        self.assertIn("50", str_repr)   # key_link

    def test_cog_hash_function(self):
        """Test standalone cog_hash function"""
        hash1 = cog_hash("test_string", 1000)
        hash2 = cog_hash("test_string", 1000)
        hash3 = cog_hash("different_string", 1000)
        
        # Same input should produce same hash
        self.assertEqual(hash1, hash2)
        # Different input may produce different hash
        # (not guaranteed, but likely with different strings)
        self.assertIsInstance(hash3, int)

    def test_index_get_with_key_chain(self):
        """Test Index.get() follows key_link chain for correct key"""
        orig_capacity = config.INDEX_CAPACITY
        config.INDEX_CAPACITY = 4
        
        logger = logging.getLogger()
        table = Table("key_chain_test", "test_table", "test_key_chain_id", config)
        store = table.store
        index = table.indexer.index_list[0]
        
        # Insert records
        rec1 = Record("chain_a", "value_a")
        rec2 = Record("chain_b", "value_b")
        
        pos1 = store.save(rec1)
        index.put(rec1.key, pos1, store)
        
        pos2 = store.save(rec2)
        index.put(rec2.key, pos2, store)
        
        # Get should find correct records
        result_a = index.get("chain_a", store)
        result_b = index.get("chain_b", store)
        
        self.assertIsNotNone(result_a)
        self.assertEqual(result_a.value, "value_a")
        self.assertIsNotNone(result_b)
        self.assertEqual(result_b.value, "value_b")
        
        table.close()
        config.INDEX_CAPACITY = orig_capacity

    def test_record_set_methods(self):
        """Test Record setter methods"""
        record = Record("key", "value")
        
        record.set_store_position(500)
        self.assertEqual(record.store_position, 500)
        
        record.set_key_link(100)
        self.assertEqual(record.key_link, 100)
        
        record.set_value_link(200)
        self.assertEqual(record.value_link, 200)
        
        record.set_value("new_value")
        self.assertEqual(record.value, "new_value")

    def test_record_set_store_position_type_error(self):
        """Test Record.set_store_position raises on non-int"""
        record = Record("key", "value")
        
        with self.assertRaises(ValueError):
            record.set_store_position("not_an_int")

    def test_store_update_record_link_type_error(self):
        """Test Store.update_record_link_inplace raises on non-int"""
        logger = logging.getLogger()
        table = Table("update_link_test", "test_table", "test_update_link_id", config)
        
        with self.assertRaises(ValueError):
            table.store.update_record_link_inplace(0, "not_an_int")
        
        table.close()

    def test_index_head_only_get(self):
        """Test Index.get_head_only for O(1) lookup"""
        logger = logging.getLogger()
        table = Table("head_only_test", "test_table", "test_head_only_id", config)
        store = table.store
        index = table.indexer.index_list[0]
        
        # Insert record with list values
        rec = Record("head_key", "head_value")
        position = store.save(rec)
        index.put(rec.key, position, store)
        
        # Get head only
        head_rec, head_pos = index.get_head_only("head_key", store)
        
        self.assertIsNotNone(head_rec)
        self.assertEqual(head_rec.key, "head_key")
        self.assertEqual(head_pos, position)
        
        # Non-existent key
        none_rec, none_pos = index.get_head_only("no_key", store)
        self.assertIsNone(none_rec)
        self.assertIsNone(none_pos)
        
        table.close()

    def test_indexer_get_head_only(self):
        """Test Indexer.get_head_only delegates to index"""
        logger = logging.getLogger()
        table = Table("indexer_head_test", "test_table", "test_indexer_head_id", config)
        
        rec = Record("indexer_head_key", "indexer_head_value")
        position = table.store.save(rec)
        table.indexer.put(rec.key, position, table.store)
        
        head_rec, head_pos = table.indexer.get_head_only("indexer_head_key", table.store)
        
        self.assertIsNotNone(head_rec)
        self.assertEqual(head_rec.key, "indexer_head_key")
        
        table.close()


if __name__ == '__main__':
    unittest.main()



from cog.torque import Graph
import unittest
import os
import shutil

DIR_NAME = "TorqueTest4"

def ordered(obj):
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in list(obj.items()))
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    else:
        return obj


class TorqueTest(unittest.TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):

        if os.path.exists("/tmp/" + DIR_NAME):
            shutil.rmtree("/tmp/" + DIR_NAME)

        if not os.path.exists("/tmp/" + DIR_NAME):
            os.mkdir("/tmp/" + DIR_NAME)

        data_dir = "test/test-data/test_func.nq"
        # choose appropriate path based on where the test is called from.
        if os.path.exists("test-data/test_func.nq"):
            data_dir = "test-data/test_func.nq"

        TorqueTest.g = Graph(graph_name="people", cog_home=DIR_NAME)
        TorqueTest.g.load_triples(data_dir, "people")
        print(">>> test setup complete.\n")

    def test_torque_func_out(self):
        expected = {'result': [{'id': 'alice'}, {'id': 'dani'}, {'id': 'greg'}]}
        actual = TorqueTest.g.v().out("score", func=lambda x: int(x) > 5).inc().all()
        self.assertTrue(expected == actual)

    def test_torque_func_out2(self):
        expected = {'result': [{'id': 'toronto'}]}
        actual = TorqueTest.g.v().out("city", func=lambda x: x.startswith("to")).all()
        self.assertTrue(expected == actual)

    def test_torque_func(self):
        expected = {'result': [{'id': 'vancouver'}]}
        actual = TorqueTest.g.v(func=lambda x: x.startswith("van")).all()
        self.assertTrue(expected == actual)

    def test_torque_func_inc(self):
        expected = {'result': [{'id': 'dani'}]}
        actual = TorqueTest.g.v().inc("city", func=lambda x: x.startswith("d")).all()
        self.assertTrue(expected == actual)

    def test_torque_func(self):
        expected = {'result': [{'id': 'edmonton'}, {'id': 'vancouver'}, {'id': 'montreal'}]}
        actual = TorqueTest.g.v().out("score", func=lambda x: int(x) > 5).inc().out("city").all()
        self.assertTrue(expected == actual)

    @classmethod
    def tearDownClass(cls):
        TorqueTest.g.close()
        shutil.rmtree("/tmp/"+DIR_NAME)
        print("*** deleted test data.")


if __name__ == '__main__':
    unittest.main()



"""
Regression test for EOFError: marshal data too short.

The bug: Store.__read_until() used RECORD_SEP (0xFD) as a delimiter to find
record boundaries.  However, marshal.dumps() can produce bytes containing
0xFD — especially when serialising floats (e.g. embeddings).  This caused
the reader to truncate records mid-payload, corrupting them on reload.

"""

import logging
from logging.config import dictConfig
import marshal
import os
import random
import shutil
import unittest

from cog.torque import Graph
from cog.core import Table, Record, Store, RECORD_SEP, UNIT_SEP
from cog import config

DIR_NAME = "TestEOFError"


class TestRecordSepInPayload(unittest.TestCase):
    """Verify records whose marshal output contains RECORD_SEP are
    round-tripped correctly through marshal/unmarshal and Store."""

    def test_marshal_unmarshal_with_fd_byte(self):
        """Record.unmarshal should survive payloads containing 0xFD."""
        # Find a key/value pair whose marshal.dumps output contains 0xFD
        random.seed(42)
        found = False
        for i in range(5000):
            key = f"entity_{i}"
            value = random.random()
            serialized = marshal.dumps((key, value))
            if RECORD_SEP in serialized:
                record = Record(key, value, key_link=100, value_type='s')
                marshalled = record.marshal()
                unmarshalled = Record.unmarshal(marshalled)
                self.assertEqual(unmarshalled.key, key)
                self.assertAlmostEqual(unmarshalled.value, value)
                found = True
                break
        self.assertTrue(found, "Could not find a key/value whose marshal output contains 0xFD")

    def test_marshal_unmarshal_list_type_with_fd_byte(self):
        """Record with value_type='l' should also survive 0xFD in payload."""
        random.seed(42)
        found = False
        for i in range(5000):
            key = f"entity_{i}"
            value = random.random()
            serialized = marshal.dumps((key, value))
            if RECORD_SEP in serialized:
                record = Record(key, value, key_link=100, value_type='l', value_link=9999)
                marshalled = record.marshal()
                unmarshalled = Record.unmarshal(marshalled)
                self.assertEqual(unmarshalled.key, key)
                self.assertAlmostEqual(unmarshalled.value, value)
                self.assertEqual(unmarshalled.value_link, 9999)
                found = True
                break
        self.assertTrue(found, "Could not find a key/value whose marshal output contains 0xFD")


class TestStoreReadWithRecordSep(unittest.TestCase):
    """Verify Store.read() correctly handles records with 0xFD in payload."""

    @classmethod
    def setUpClass(cls):
        path = "/tmp/" + DIR_NAME + "/test_table/"
        if not os.path.exists(path):
            os.makedirs(path)
        config.CUSTOM_COG_DB_PATH = "/tmp/" + DIR_NAME

    @classmethod
    def tearDownClass(cls):
        if os.path.exists("/tmp/" + DIR_NAME):
            shutil.rmtree("/tmp/" + DIR_NAME)

    def test_store_roundtrip_with_fd_in_payload(self):
        """Write records containing 0xFD in their serialized payload,
        then read them back via Store.read() and verify correctness."""
        dictConfig(config.logging_config)
        logger = logging.getLogger()

        table = Table("testdb_eof", "test_table", "test_eof_instance", config, logger)
        store = table.store
        index = table.indexer.index_list[0]

        # Collect records whose marshal.dumps contains 0xFD
        random.seed(42)
        problematic_records = []
        for i in range(1000):
            key = f"node_{i}"
            value = random.random()
            serialized = marshal.dumps((key, value))
            if RECORD_SEP in serialized:
                problematic_records.append((key, value))
            if len(problematic_records) >= 10:
                break

        self.assertGreater(len(problematic_records), 0,
                           "Need at least one record with 0xFD in payload")

        # Write them through the Store
        positions = []
        for key, value in problematic_records:
            record = Record(key, value)
            pos = store.save(record)
            index.put(key, pos, store)
            positions.append(pos)

        # Read back and verify
        for (key, value), pos in zip(problematic_records, positions):
            raw = store.read(pos)
            self.assertIsNotNone(raw, f"Store.read returned None for key={key}")
            recovered = Record.unmarshal(raw)
            self.assertEqual(recovered.key, key)
            self.assertAlmostEqual(recovered.value, value)

        table.close()


class TestGraphReopenWithEmbeddings(unittest.TestCase):
    """End-to-end: create a graph with embeddings (floats), close it,
    reopen it, and verify data survives — the exact scenario that
    triggered the original EOFError."""

    @classmethod
    def setUpClass(cls):
        cls.cog_home = "/tmp/" + DIR_NAME + "_graph"
        if os.path.exists(cls.cog_home):
            shutil.rmtree(cls.cog_home)

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(cls.cog_home):
            shutil.rmtree(cls.cog_home)

    def test_graph_reopen_with_embeddings(self):
        """Graph with float embeddings should survive close + reopen."""
        g = Graph("EOFTest", cog_home=self.cog_home)

        # Insert triples
        triples = []
        for i in range(50):
            triples.append((f"entity_{i}", "relates_to", f"target_{i}"))
        g.put_batch(triples)

        # Insert embeddings (floats are the main source of 0xFD bytes)
        random.seed(42)
        entities_with_embeddings = []
        for i in range(50):
            name = f"entity_{i}"
            vec = [random.random() for _ in range(16)]
            g.put_embedding(name, vec)
            entities_with_embeddings.append((name, vec))

        g.close()

        # Reopen — this is where the old code would crash with
        # EOFError: marshal data too short
        g2 = Graph("EOFTest", cog_home=self.cog_home)

        # Verify graph traversal still works
        result = g2.v("entity_0").out().all()
        self.assertIsNotNone(result)
        self.assertIn("result", result)
        ids = [r["id"] for r in result["result"]]
        self.assertIn("target_0", ids)

        # Verify embeddings still work
        nearest = g2.v().k_nearest("entity_0", k=3).all()
        self.assertIsNotNone(nearest)
        self.assertIn("result", nearest)
        self.assertGreater(len(nearest["result"]), 0)

        g2.close()


class TestTruncatedStore(unittest.TestCase):
    """Verify that __read_record handles truncated store files gracefully
    (returns None instead of raising IndexError or corrupt marshal)."""

    @classmethod
    def setUpClass(cls):
        cls.dir = "/tmp/TestTruncatedStore"
        if os.path.exists(cls.dir):
            shutil.rmtree(cls.dir)
        os.makedirs(cls.dir + "/test_table/", exist_ok=True)
        config.CUSTOM_COG_DB_PATH = cls.dir
        dictConfig(config.logging_config)

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(cls.dir):
            shutil.rmtree(cls.dir)

    def test_truncated_header_returns_none(self):
        """If the store file is truncated mid-header (<18 bytes),
        Store.read() should return None — not raise IndexError."""
        logger = logging.getLogger()
        table = Table("trunc_hdr", "test_table", "trunc_inst", config, logger)
        store = table.store

        # Write a valid record
        rec = Record("key_trunc", "value_trunc")
        pos = store.save(rec)

        # Truncate the file to only 10 bytes (inside the 18-byte header)
        store.store_file.flush()
        store_path = store.store
        table.close()

        with open(store_path, 'r+b') as f:
            f.truncate(10)

        # Reopen and try to read from position 0
        table2 = Table("trunc_hdr", "test_table", "trunc_inst", config, logger)
        result = table2.store.read(0)
        self.assertIsNone(result)
        table2.close()

    def test_truncated_payload_returns_none(self):
        """If the store file is truncated mid-payload,
        Store.read() should return None — not corrupt marshal."""
        logger = logging.getLogger()
        table = Table("trunc_pay", "test_table", "trunc_inst2", config, logger)
        store = table.store

        # Write a valid record
        rec = Record("key_payload", "a_longer_value_for_testing")
        pos = store.save(rec)

        store.store_file.flush()
        store_path = store.store

        # Find out how big the header+length field is, then truncate inside payload
        # Header = 16 (key_link) + 1 (tombstone) + 1 (value_type) = 18
        # Then variable-len digits + UNIT_SEP. We'll truncate at header + 5 bytes
        # (enough for len digits + UNIT_SEP + a few payload bytes, but not all).
        table.close()

        with open(store_path, 'r+b') as f:
            f.truncate(25)  # well past header, but before full payload

        table2 = Table("trunc_pay", "test_table", "trunc_inst2", config, logger)
        result = table2.store.read(0)
        self.assertIsNone(result)
        table2.close()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Test hash collision handling in CogDB.

This test uses a very small INDEX_CAPACITY to force collisions, then verifies:
1. All records can be retrieved correctly via get()
2. Scanner returns all records (follows key_link chains)
3. Embeddings with collisions work correctly
"""
import unittest
import os
import shutil
import sys

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from cog.torque import Graph
from cog.database import Cog
from cog.core import Record
from cog import config

# Save original config
ORIGINAL_INDEX_CAPACITY = config.INDEX_CAPACITY

# Test directory
TEST_DIR = "/tmp/CogCollisionTest"


class TestHashCollisions(unittest.TestCase):
    """Test that hash collisions are handled correctly."""

    @classmethod
    def setUpClass(cls):
        """Set up test with very small INDEX_CAPACITY to force collisions."""
        if os.path.exists(TEST_DIR):
            shutil.rmtree(TEST_DIR)
        os.makedirs(TEST_DIR)
        
        # Use very small capacity to guarantee collisions
        # With capacity=101 and 500 items, average ~5 collisions per bucket
        config.INDEX_CAPACITY = 101  # Small prime number

    @classmethod
    def tearDownClass(cls):
        """Restore original config and clean up."""
        config.INDEX_CAPACITY = ORIGINAL_INDEX_CAPACITY
        if os.path.exists(TEST_DIR):
            shutil.rmtree(TEST_DIR)

    def setUp(self):
        """Clean test directory before each test."""
        test_path = os.path.join(TEST_DIR, "test_data")
        if os.path.exists(test_path):
            shutil.rmtree(test_path)

    # =========================================================================
    # Basic key-value collision tests
    # =========================================================================

    def test_basic_collision_handling(self):
        """Test that all records are retrievable despite hash collisions."""
        cog = Cog()
        cog.create_or_load_namespace("collision_test")
        cog.create_table("test_table", "collision_test")

        # Insert 500 records with capacity=101, guaranteeing many collisions
        num_records = 500
        for i in range(num_records):
            key = f"key_{i}"
            value = f"value_{i}"
            cog.put(Record(key, value))

        # Verify all records can be retrieved
        retrieved = 0
        for i in range(num_records):
            key = f"key_{i}"
            expected_value = f"value_{i}"
            record = cog.get(key)
            self.assertIsNotNone(record, f"Record with key '{key}' not found")
            self.assertEqual(record.value, expected_value, f"Wrong value for key '{key}'")
            retrieved += 1

        self.assertEqual(retrieved, num_records, "Not all records were retrieved")
        cog.close()

    def test_scanner_returns_all_collided_records(self):
        """Test that scanner returns all records including those with hash collisions."""
        cog = Cog()
        cog.create_or_load_namespace("scanner_test")
        cog.create_table("test_table", "scanner_test")

        # Insert records
        num_records = 500
        inserted_keys = set()
        for i in range(num_records):
            key = f"key_{i}"
            value = f"value_{i}"
            cog.put(Record(key, value))
            inserted_keys.add(key)

        # Scan and collect all keys
        scanned_keys = set()
        for record in cog.scanner():
            scanned_keys.add(record.key)

        # All inserted keys should be found by scanner
        self.assertEqual(len(scanned_keys), num_records, 
                        f"Scanner returned {len(scanned_keys)} records, expected {num_records}")
        self.assertEqual(inserted_keys, scanned_keys, 
                        "Scanner did not return all inserted keys")

        cog.close()

    def test_update_with_collisions(self):
        """Test that updates work correctly with hash collisions."""
        cog = Cog()
        cog.create_or_load_namespace("update_test")
        cog.create_table("test_table", "update_test")

        # Insert initial records
        num_records = 200
        for i in range(num_records):
            cog.put(Record(f"key_{i}", f"original_{i}"))

        # Update all records
        for i in range(num_records):
            cog.put(Record(f"key_{i}", f"updated_{i}"))

        # Verify updates
        for i in range(num_records):
            record = cog.get(f"key_{i}")
            self.assertIsNotNone(record)
            self.assertEqual(record.value, f"updated_{i}")

        cog.close()

    # =========================================================================
    # Graph collision tests
    # =========================================================================

    def test_graph_collision_handling(self):
        """Test that graph operations work correctly with hash collisions."""
        g = Graph("collision_graph", cog_home="CollisionTest", cog_path_prefix=TEST_DIR)

        # Add many edges to force collisions
        num_nodes = 300
        for i in range(num_nodes):
            g.put(f"node_{i}", "connects", f"node_{(i+1) % num_nodes}")

        # Verify all nodes are queryable
        for i in range(0, num_nodes, 10):  # Sample every 10th node
            result = g.v(f"node_{i}").out("connects").all()
            self.assertEqual(len(result['result']), 1)
            expected = f"node_{(i+1) % num_nodes}"
            self.assertEqual(result['result'][0]['id'], expected)

        # Verify scan returns all nodes (may include edge targets)
        scan_result = g.scan(limit=num_nodes + 100)
        self.assertGreaterEqual(len(scan_result['result']), num_nodes)

        g.close()

    # =========================================================================
    # Embedding collision tests
    # =========================================================================

    def test_embedding_collision_handling(self):
        """Test that embeddings work correctly with hash collisions."""
        g = Graph("embed_collision", cog_home="EmbedCollisionTest", cog_path_prefix=TEST_DIR)

        # Generate embeddings for many words
        num_embeddings = 500
        embedding_dim = 10
        
        # Store embeddings
        for i in range(num_embeddings):
            word = f"word_{i}"
            # Deterministic embedding based on word index
            embedding = [float(i + j) / num_embeddings for j in range(embedding_dim)]
            g.put_embedding(word, embedding)

        # Verify all embeddings can be retrieved
        retrieved = 0
        for i in range(num_embeddings):
            word = f"word_{i}"
            embedding = g.get_embedding(word)
            self.assertIsNotNone(embedding, f"Embedding for '{word}' not found")
            self.assertEqual(len(embedding), embedding_dim)
            # Verify first value matches expected
            expected_first = float(i) / num_embeddings
            self.assertAlmostEqual(embedding[0], expected_first, places=5)
            retrieved += 1

        self.assertEqual(retrieved, num_embeddings)

        # Verify embedding_stats returns correct count
        stats = g.embedding_stats()
        self.assertEqual(stats['count'], num_embeddings)
        self.assertEqual(stats['dimensions'], embedding_dim)

        g.close()

    def test_k_nearest_with_collisions(self):
        """Test that k_nearest works correctly when embeddings have hash collisions."""
        g = Graph("knn_collision", cog_home="KNNCollisionTest", cog_path_prefix=TEST_DIR)

        # Create embeddings in clusters to test similarity search
        # Cluster 1: words 0-99 have similar embeddings
        # Cluster 2: words 100-199 have similar embeddings
        # etc.
        num_embeddings = 300
        embedding_dim = 10

        for i in range(num_embeddings):
            word = f"word_{i}"
            cluster = i // 100
            # Base vector for cluster + small noise
            embedding = [float(cluster) + (i % 100) * 0.001 + j * 0.01 
                        for j in range(embedding_dim)]
            g.put_embedding(word, embedding)

        # Find k-nearest to word_50 (cluster 0)
        # Note: Don't call g.v() first - that returns empty list (no graph vertices)
        # Instead, call k_nearest directly which scans the embedding table
        result = g.k_nearest("word_50", k=5).all()
        
        # All nearest neighbors should be from cluster 0 (words 0-99)
        self.assertEqual(len(result['result']), 5)
        for item in result['result']:
            word_id = item['id']
            word_num = int(word_id.split('_')[1])
            self.assertTrue(0 <= word_num < 100, 
                          f"Expected cluster 0 word, got {word_id}")

        g.close()

    def test_batch_embedding_with_collisions(self):
        """Test that batch embedding insertion works with collisions."""
        g = Graph("batch_collision", cog_home="BatchCollisionTest", cog_path_prefix=TEST_DIR)

        # Prepare batch data
        num_embeddings = 400
        embedding_dim = 10
        batch = []
        
        for i in range(num_embeddings):
            word = f"batchword_{i}"
            embedding = [float(i * j) / 1000 for j in range(embedding_dim)]
            batch.append((word, embedding))

        # Batch insert
        g.put_embeddings_batch(batch)

        # Verify all are retrievable
        stats = g.embedding_stats()
        self.assertEqual(stats['count'], num_embeddings)

        # Spot check some embeddings
        for i in [0, 100, 200, 399]:
            word = f"batchword_{i}"
            embedding = g.get_embedding(word)
            self.assertIsNotNone(embedding, f"Embedding for '{word}' not found")
            self.assertEqual(len(embedding), embedding_dim)

        g.close()


if __name__ == '__main__':
    unittest.main()



script_part1 =r"""
<!DOCTYPE html>
<html lang="en">
  <head>
    <title>Cog Graph</title>
    <style type="text/css">
       body {
        padding: 0;
        margin: 0;
        width: 100%;!important; 
        height: 100%;!important; 
      }

      #cog-graph-view {
        width: 700px;
        height: 700px;
      }
    </style>
"""

graph_lib_src = r"""

    <script
      type="text/javascript"
      src="{js_src}"
    ></script>
  </head>  
"""

graph_template = r""" 
  <body>
    <div id="cog-graph-view"></div>

    <script type="text/javascript">

    var graphData ={plot_data_insert} """

script_part2 = r"""

    var nodes = new vis.DataSet();
    var edges = new vis.DataSet();

    for (var i = 0; i < graphData.nodes.length; i++) {
        var n = graphData.nodes[i];
        nodes.update({ id: n.id, label: n.id });
    }

    for (var i = 0; i < graphData.links.length; i++) {
        var link = graphData.links[i];
        edges.update({
            id: link.source + "-" + link.label + "-" + link.target,
            from: link.source,
            to: link.target,
            label: link.label || ""
        });
    }

    var container = document.getElementById("cog-graph-view");
    var data = {
        nodes: nodes,
        edges: edges,
    };
    var options = {
        nodes: {
            font: {
                size: 20,
                color: "black"
            },
            color: "#46944f",
            shape: "dot",
            widthConstraint: 200,

        },
        edges: {
            font: { size: 12, color: "#555555", strokeWidth: 0, align: "middle" },
            scaling: {
                label: true,
            },
            shadow: true,
            smooth: true,
            arrows: { to: {enabled: true}}
        },
        physics: {
            barnesHut: {
                gravitationalConstant: -30000
            },
            stabilization: {
                iterations: 1000
            },
        }

    };
    var network = new vis.Network(container, data, options);
    </script>
  </body>
</html>

"""


# share.py - WebSocket share client for CogDB Studio
"""
Share client that connects a local g.serve() instance to the CogDB relay,
"""
import json
import logging
import re
import threading
import urllib.request
import urllib.error

logger = logging.getLogger("cog.share")


def _validate_path(path):
    """
    Validate that path is safe to forward to local server.
    """
    if not path:
        return False
    # Path must start with /
    if not path.startswith('/'):
        return False
    # Path must not contain @ (urllib can interpret this as host)
    if '@' in path:
        return False
    # Path must be a valid URL path (alphanumeric, /, -, _, ., ?, &, =, etc.)
    if not re.match(r'^[a-zA-Z0-9/_\-\.\?&=%]+$', path):
        return False
    return True


def _http_request(method, url, body=None, timeout=25, extra_headers=None):
    data = body.encode('utf-8') if body else None
    req = urllib.request.Request(url, data=data, method=method)
    req.add_header('Content-Type', 'application/json')
    if extra_headers:
        for key, value in extra_headers.items():
            req.add_header(key, value)
    
    try:
        with urllib.request.urlopen(req, timeout=timeout) as resp:
            return resp.status, resp.read().decode('utf-8'), resp.headers.get('Content-Type', 'application/json')
    except urllib.error.HTTPError as e:
        return e.code, e.read().decode('utf-8'), e.headers.get('Content-Type', 'application/json')
    except Exception as e:
        return 502, str(e), 'text/plain'


class ShareInfo:
    """Holds share connection information."""
    def __init__(self):
        self.url = None
        self.session_id = None
        self._ready = threading.Event()
        self._error = None
    
    def wait(self, timeout=10):
        """Wait for share URL to be available. Returns URL or raises error."""
        if self._ready.wait(timeout):
            if self._error:
                raise ConnectionError(self._error)
            return self.url
        raise TimeoutError("Timed out waiting for share connection")


def start_share(local_port, local_host="127.0.0.1", relay_url=None):
    """
    Connect to CogDB relay and share requests to local g.serve().
    
    Args:
        local_port: Local HTTP port where g.serve() is running
        local_host: Local host where g.serve() is bound (default 127.0.0.1)
        relay_url: WebSocket URL for the relay server (uses config.RELAY_URL if None)
    
    Returns:
        ShareInfo: Object with .url property and .wait() method
    
    Raises:
        ImportError: If websocket-client is not installed
        RuntimeError: If sharing is disabled (RELAY_URL is None in config)
    """
    try:
        import websocket
    except ImportError:
        raise ImportError(
            "Share requires 'websocket-client' package. "
            "Install with: pip install websocket-client"
        )
    
    # Get relay URL from config if not provided
    if relay_url is None:
        from cog import config as cfg
        relay_url = cfg.RELAY_URL
    
    # Check if sharing is disabled
    if relay_url is None:
        raise RuntimeError(
            "CogDB sharing is disabled. Set RELAY_URL in cog/config.py to enable sharing."
        )
    
    import certifi
    sslopt = {"ca_certs": certifi.where()}
    
    # Normalize host
    forward_host = "127.0.0.1" if local_host == "0.0.0.0" else local_host
    
    info = ShareInfo()
    
    def run():
        try:
            ws = websocket.WebSocket(sslopt=sslopt)
            ws.connect(relay_url)
            
            # Wait for session message
            msg = json.loads(ws.recv())
            if msg.get("type") != "conn.session":
                info._error = "Expected conn.session message from relay"
                info._ready.set()
                return
            
            # Extract share URL from session
            info.session_id = msg.get("session")
            info.url = msg.get("share_url")  # The relay provides the public URL
            logger.debug(f"Relay session established: session={info.session_id}, url={info.url}")
            info._ready.set()
            
            # Forward requests from relay to local server
            while True:
                msg = {}
                try:
                    msg = json.loads(ws.recv())
                    msg_type = msg.get("type")
                    
                    if msg_type == "share.request":
                        path = msg.get("path", "")
                        
                        if not _validate_path(path):
                            logger.warning(f"Rejected invalid path from relay: {path}")
                            ws.send(json.dumps({
                                "type": "share.response",
                                "req_id": msg["req_id"],
                                "status": 400,
                                "body": "Invalid path",
                                "content_type": "text/plain"
                            }))
                            continue
                        
                        # Pass headers so server generates correct links for remote access
                        extra_headers = {
                            "X-Share-Url": info.url or ""
                        } if info.url else None
                        status, body, content_type = _http_request(
                            method=msg["method"],
                            url=f"http://{forward_host}:{local_port}{path}",
                            body=msg.get("body"),
                            extra_headers=extra_headers
                        )
                        ws.send(json.dumps({
                            "type": "share.response",
                            "req_id": msg["req_id"],
                            "status": status,
                            "body": body,
                            "content_type": content_type
                        }))
                    
                    elif msg_type == "conn.ping":
                        ws.send(json.dumps({"type": "conn.pong"}))
                    
                except Exception as e:
                    if msg.get("req_id"):
                        try:
                            ws.send(json.dumps({
                                "type": "share.response",
                                "req_id": msg["req_id"],
                                "status": 502,
                                "body": f"Share error: {str(e)}",
                                "content_type": "text/plain"
                            }))
                        except Exception:
                            break
        except Exception as e:
            info._error = str(e)
            info._ready.set()
    
    t = threading.Thread(target=run, daemon=True)
    t.start()
    return info



"""
Tests for batch mode operations in CogDB.
Tests both correctness and performance improvement.
"""
import os
import shutil
import timeit
import unittest
import logging
from logging.config import dictConfig

from cog import config
from cog.core import Table, Record
from cog.database import Cog
from cog.torque import Graph

DIR_NAME = "TestBatchMode"


def ensure_namespace_dir(namespace):
    """Helper to create namespace directory if it doesn't exist"""
    path = config.cog_data_dir(namespace)
    if not os.path.exists(path):
        os.makedirs(path)


class TestStoreBatchMode(unittest.TestCase):
    """Tests for Store.begin_batch() and Store.end_batch()"""

    @classmethod
    def setUpClass(cls):
        # Clean up any existing test data
        base_path = "/tmp/" + DIR_NAME
        if os.path.exists(base_path):
            shutil.rmtree(base_path)
        os.makedirs(base_path)
        config.CUSTOM_COG_DB_PATH = "/tmp/" + DIR_NAME
        dictConfig(config.logging_config)

    def test_batch_mode_saves_correctly(self):
        """Test that batch mode saves records correctly"""
        logger = logging.getLogger()
        
        # Create namespace directory
        ensure_namespace_dir("test_batch")
        
        table = Table("testdb", "test_batch", "test_batch_id", config, logger)
        store = table.store
        indexer = table.indexer

        # Enable batch mode
        store.begin_batch()
        
        records = []
        for i in range(100):
            r = Record(f"key_{i}", f"value_{i}")
            records.append(r)
            position = store.save(r)
            indexer.put(r.key, position, store)
        
        # End batch mode (flush)
        store.end_batch()
        
        # Verify all records are readable
        for r in records:
            retrieved = indexer.get(r.key, store)
            self.assertIsNotNone(retrieved, f"Record {r.key} should exist")
            self.assertEqual(r.key, retrieved.key)
            self.assertEqual(r.value, retrieved.value)
        
        table.close()

    def test_batch_mode_closes_safely(self):
        """Test that table close flushes pending writes if still in batch mode"""
        logger = logging.getLogger()
        
        # Create namespace directory
        ensure_namespace_dir("test_batch_close")
        
        table = Table("testdb", "test_batch_close", "test_batch_close_id", config, logger)
        store = table.store
        indexer = table.indexer

        # Enable batch mode but don't call end_batch()
        store.begin_batch()
        
        r = Record("orphan_key", "orphan_value")
        position = store.save(r)
        indexer.put(r.key, position, store)
        
        # Close without end_batch - should still flush
        table.close()
        
        # Reopen and verify
        table2 = Table("testdb", "test_batch_close", "test_batch_close_id", config, logger)
        retrieved = table2.indexer.get("orphan_key", table2.store)
        self.assertIsNotNone(retrieved, "Record should exist after close")
        self.assertEqual("orphan_value", retrieved.value)
        table2.close()

    def test_batch_mode_performance_improvement(self):
        """Test that batch mode provides performance improvement on larger datasets"""
        logger = logging.getLogger()
        
        # Use larger dataset to see measurable difference
        num_records = 3000
        
        # Create namespace directories
        ensure_namespace_dir("test_perf_no_batch")
        ensure_namespace_dir("test_perf_batch")
        
        # Test without batch mode
        table1 = Table("testdb", "test_perf_no_batch", "test_perf_nb_id", config, logger)
        
        start = timeit.default_timer()
        for i in range(num_records):
            r = Record(f"key_nb_{i}", f"value_nb_{i}")
            position = table1.store.save(r)
            table1.indexer.put(r.key, position, table1.store)
        time_no_batch = timeit.default_timer() - start
        table1.close()
        
        # Test with batch mode
        table2 = Table("testdb", "test_perf_batch", "test_perf_b_id", config, logger)
        table2.store.begin_batch()
        
        start = timeit.default_timer()
        for i in range(num_records):
            r = Record(f"key_b_{i}", f"value_b_{i}")
            position = table2.store.save(r)
            table2.indexer.put(r.key, position, table2.store)
        table2.store.end_batch()
        time_with_batch = timeit.default_timer() - start
        table2.close()
        
        # Report performance (actual speedup varies by OS/disk)
        speedup = time_no_batch / time_with_batch if time_with_batch > 0 else 1.0
        print(f"\n=== Store Batch Mode Performance ===")
        print(f"Records: {num_records}")
        print(f"Without batch: {time_no_batch:.4f}s ({num_records/time_no_batch:.0f} ops/s)")
        print(f"With batch: {time_with_batch:.4f}s ({num_records/time_with_batch:.0f} ops/s)")
        print(f"Speedup: {speedup:.2f}x")
        
        # Batch mode should complete without errors - performance varies by system
        self.assertTrue(True, "Batch mode completed successfully")

    @classmethod
    def tearDownClass(cls):
        # Clean up after this test class
        pass


class TestCogBatchMode(unittest.TestCase):
    """Tests for Cog.begin_batch() and Cog.end_batch()"""

    @classmethod
    def setUpClass(cls):
        config.CUSTOM_COG_DB_PATH = "/tmp/" + DIR_NAME
        dictConfig(config.logging_config)

    def test_cog_batch_mode(self):
        """Test batch mode at Cog level"""
        cog = Cog()
        cog.create_or_load_namespace("test_cog_batch")
        cog.create_table("batch_table", "test_cog_batch")
        
        cog.begin_batch()
        
        for i in range(50):
            cog.put(Record(f"cog_key_{i}", f"cog_value_{i}"))
        
        cog.end_batch()
        
        # Verify data
        for i in range(50):
            result = cog.get(f"cog_key_{i}")
            self.assertIsNotNone(result)
            self.assertEqual(f"cog_value_{i}", result.value)
        
        cog.close()


class TestGraphPutBatch(unittest.TestCase):
    """Tests for Graph.put_batch()"""

    @classmethod
    def setUpClass(cls):
        config.CUSTOM_COG_DB_PATH = "/tmp/" + DIR_NAME
        dictConfig(config.logging_config)

    def test_put_batch_correctness(self):
        """Test that put_batch inserts all triples correctly"""
        g = Graph("test_put_batch")
        
        triples = [
            ("alice", "follows", "bob"),
            ("bob", "follows", "charlie"),
            ("charlie", "follows", "alice"),
            ("alice", "likes", "pizza"),
            ("bob", "likes", "tacos"),
        ]
        
        g.put_batch(triples)
        
        # Verify edges
        result = g.v("alice").out("follows").all()
        self.assertEqual(1, len(result['result']))
        self.assertEqual("bob", result['result'][0]['id'])
        
        result = g.v("bob").out("follows").all()
        self.assertEqual(1, len(result['result']))
        self.assertEqual("charlie", result['result'][0]['id'])
        
        result = g.v("charlie").out("follows").all()
        self.assertEqual(1, len(result['result']))
        self.assertEqual("alice", result['result'][0]['id'])
        
        result = g.v("alice").out("likes").all()
        self.assertEqual(1, len(result['result']))
        self.assertEqual("pizza", result['result'][0]['id'])
        
        g.close()

    def test_put_batch_multiple_edges_same_predicate(self):
        """Test put_batch with multiple edges from same vertex"""
        g = Graph("test_put_batch_multi")
        
        triples = [
            ("alice", "follows", "bob"),
            ("alice", "follows", "charlie"),
            ("alice", "follows", "dave"),
        ]
        
        g.put_batch(triples)
        
        result = g.v("alice").out("follows").all()
        self.assertEqual(3, len(result['result']))
        
        ids = {r['id'] for r in result['result']}
        self.assertEqual({"bob", "charlie", "dave"}, ids)
        
        g.close()

    def test_put_batch_performance(self):
        """Test that put_batch works correctly and report performance"""
        num_edges = 1000
        
        # Generate test data
        triples = [(f"user_{i}", "follows", f"user_{i+1}") for i in range(num_edges)]
        
        # Test individual puts
        g1 = Graph("test_perf_individual")
        start = timeit.default_timer()
        for v1, pred, v2 in triples:
            g1.put(v1, pred, v2)
        time_individual = timeit.default_timer() - start
        g1.close()
        
        # Test put_batch
        g2 = Graph("test_perf_batch_graph")
        start = timeit.default_timer()
        g2.put_batch(triples)
        time_batch = timeit.default_timer() - start
        g2.close()
        
        speedup = time_individual / time_batch if time_batch > 0 else 1.0
        print(f"\n=== Graph.put_batch Performance ===")
        print(f"Edges: {num_edges}")
        print(f"Individual puts: {time_individual:.4f}s ({num_edges/time_individual:.0f} edges/s)")
        print(f"put_batch: {time_batch:.4f}s ({num_edges/time_batch:.0f} edges/s)")
        print(f"Speedup: {speedup:.2f}x")
        
        # Verify data was inserted correctly
        g3 = Graph("test_perf_batch_graph")
        result = g3.v("user_0").out("follows").all()
        self.assertEqual(1, len(result['result']))
        g3.close()

    def test_put_batch_empty_list(self):
        """Test put_batch with empty list doesn't crash"""
        g = Graph("test_put_batch_empty")
        g.put_batch([])  # Should not raise
        g.close()

    def test_put_batch_error_recovery(self):
        """Test that batch mode is properly ended even on error"""
        g = Graph("test_put_batch_error")
        
        # First batch should work
        g.put_batch([("a", "rel", "b")])
        
        # Verify insert worked
        result = g.v("a").out("rel").all()
        self.assertEqual(1, len(result['result']))
        
        g.close()


class TestBatchModeIntegration(unittest.TestCase):
    """Integration tests for batch mode across all layers"""

    @classmethod
    def setUpClass(cls):
        config.CUSTOM_COG_DB_PATH = "/tmp/" + DIR_NAME
        dictConfig(config.logging_config)

    def test_large_graph_insertion(self):
        """Test inserting a large graph using batch mode"""
        g = Graph("test_large_graph")
        
        # Create a larger graph
        num_nodes = 100
        triples = []
        
        for i in range(num_nodes):
            for j in range(i + 1, min(i + 5, num_nodes)):  # Connect to next 4 nodes
                triples.append((f"node_{i}", "connects", f"node_{j}"))
        
        print(f"\n=== Large Graph Test ===")
        print(f"Nodes: {num_nodes}")
        print(f"Edges: {len(triples)}")
        
        start = timeit.default_timer()
        g.put_batch(triples)
        elapsed = timeit.default_timer() - start
        
        print(f"Insertion time: {elapsed:.4f}s")
        print(f"Edges per second: {len(triples)/elapsed:.0f}")
        
        # Verify some random edges
        result = g.v("node_0").out("connects").all()
        self.assertGreater(len(result['result']), 0)
        
        result = g.v("node_50").out("connects").all()
        self.assertGreater(len(result['result']), 0)
        
        g.close()


def tearDownModule():
    """Clean up all test data"""
    if os.path.exists("/tmp/" + DIR_NAME):
        shutil.rmtree("/tmp/" + DIR_NAME)
    print("\n*** Cleaned up test data.")


if __name__ == '__main__':
    unittest.main(verbosity=2)



from cog.core import Record
from cog.core import Index
from cog.core import Store
from cog.core import TableMeta
from cog import config
import logging
import os
import shutil
from logging.config import dictConfig
import random
import string

import unittest


DIR_NAME = "TestIndexTemp"


class TestIndex2(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        path = "/tmp/"+DIR_NAME+"/test_table"
        if not os.path.exists(path):
            os.makedirs(path)
        config.CUSTOM_COG_DB_PATH = "/tmp/"+DIR_NAME
        print("*** created: "+path)

    def test_put_get(self):

        dictConfig(config.logging_config)
        logger = logging.getLogger()
        tablemeta = TableMeta("testdb", "test_table", "test_xcvzdfsadx", None)
        store = Store(tablemeta, config, logger)
        index = Index(tablemeta, config, logger, 0)
        test_size = 30
        for i in range(test_size):
            # Use deterministic keys to avoid hash collision flakiness
            key = f"idx_key_{i:010d}"
            value = f"idx_value_{i:0100d}"
            expected_data = Record(key, value)
            position=store.save(expected_data)
            index.put(expected_data.key,position,store)
            returned_data=index.get(expected_data.key, store)
            self.assertTrue(expected_data.is_equal_val(returned_data))

        c = 0
        scanner = index.scanner(store)
        for r in scanner:
            print(r)
            c += 1
        self.assertEqual(test_size, c)

        index.close()
        store.close()

    @classmethod
    def tearDownClass(cls):
        path = "/tmp/"+DIR_NAME
        shutil.rmtree(path)
        print("*** deleted test data." + path)


if __name__ == '__main__':
    unittest.main()



COG_PATH_PREFIX = "/tmp"
COG_ROOT = "cog_root"
COG_HOME = "cog-test"
COG_SYS_DIR = "sys"
COG_SYS_FILE = "cogsys.c"
COG_DEFAULT_NAMESPACE = "default"
VIEWS = 'views'
STORE="-store-"
INDEX="-index-"
INDEX_BLOCK_LEN = 32
INDEX_CAPACITY = 100003 # must be a prime number
STORE_READ_BUFFER_SIZE = 512
LEVEL_2_CACHE_SIZE = 100000

''' TORQUE '''
GRAPH_NODE_SET_TABLE_NAME = 'TOR_NODE_SET'
GRAPH_EDGE_SET_TABLE_NAME = 'TOR_EDGE_SET'
EMBEDDING_SET_TABLE_NAME = 'EMBEDDING_SET'

''' CUSTOM COG DB PATH '''
CUSTOM_COG_DB_PATH = None

''' SHARE/TUNNEL RELAY '''
# Set to None to disable the share feature entirely
# When None, calling serve(share=True) will raise an error
RELAY_URL = "wss://s.cogdb.io/register"


def cog_db_path():
    if CUSTOM_COG_DB_PATH:
        return CUSTOM_COG_DB_PATH
    else:
        return "/".join([COG_PATH_PREFIX, COG_HOME])


def cog_context():
    return [cog_db_path(), COG_SYS_DIR, COG_SYS_FILE]


def cog_instance_sys_file():
    return "/".join(cog_context())


def cog_instance_sys_dir():
    return "/".join(cog_context()[0:-1])


def cog_views_dir():
    return "/".join([cog_db_path(), VIEWS])


def cog_data_dir(db_name):
    return "/".join(cog_context()[0:-2]+[db_name])


def cog_index(db_name, table_name, instance_id, index_id):
    return "/".join(cog_context()[0:-2]+[db_name,table_name+INDEX+instance_id+"-"+str(index_id)])


def get_table_name(index_file_name):
    return index_file_name.split(INDEX)[0]


def index_id(index_name):
    return int(index_name.split("-")[-1])


def cog_store(db_name, table_name, instance_id):
    return "/".join(cog_context()[0:-2]+[db_name, table_name+STORE+instance_id])


import logging

logging_config = dict(
    version = 1,
    formatters = {
        'f': {'format':
              '%(asctime)s %(name)-12s %(levelname)-8s [%(filename)s:%(lineno)s - %(funcName)10s()] %(message)s'}
        },
    handlers = {
        'h': {'class': 'logging.StreamHandler',
              'formatter': 'f'
              }
        },
    root = {
        'handlers': ['h'],
        'level': logging.WARN,
        }
)


"""
Tests for CogDB Share/Tunnel Feature

Tests for the share.py module and related template/server changes
that enable sharing graphs via the CogDB Studio relay.
"""

import unittest
import os
import shutil
import time
import json
import urllib.request


DIR_NAME = "ShareTest"


class TestShareInfo(unittest.TestCase):
    """Test the ShareInfo class functionality."""
    
    def test_shareinfo_wait_timeout(self):
        """ShareInfo.wait() raises TimeoutError when not ready."""
        from cog.share import ShareInfo
        
        info = ShareInfo()
        # Should timeout since _ready is never set
        with self.assertRaises(TimeoutError):
            info.wait(timeout=0.1)
    
    def test_shareinfo_wait_with_error(self):
        """ShareInfo.wait() raises ConnectionError when error is set."""
        from cog.share import ShareInfo
        
        info = ShareInfo()
        info._error = "Test connection error"
        info._ready.set()
        
        with self.assertRaises(ConnectionError) as ctx:
            info.wait(timeout=1)
        
        self.assertIn("Test connection error", str(ctx.exception))
    
    def test_shareinfo_wait_success(self):
        """ShareInfo.wait() returns URL when ready."""
        from cog.share import ShareInfo
        
        info = ShareInfo()
        info.url = "https://test123.s.cogdb.io/"
        info.session_id = "test123"
        info._ready.set()
        
        result = info.wait(timeout=1)
        self.assertEqual(result, "https://test123.s.cogdb.io/")


class TestPathValidation(unittest.TestCase):
    """Test SSRF prevention via path validation."""
    
    def test_validate_path_empty(self):
        """Empty path is rejected."""
        from cog.share import _validate_path
        
        self.assertFalse(_validate_path(""))
        self.assertFalse(_validate_path(None))
    
    def test_validate_path_must_start_with_slash(self):
        """Path must start with /."""
        from cog.share import _validate_path
        
        self.assertFalse(_validate_path("graph/query"))
        self.assertTrue(_validate_path("/graph/query"))
    
    def test_validate_path_rejects_at_symbol(self):
        """Path with @ is rejected (SSRF prevention)."""
        from cog.share import _validate_path
        
        # The @ symbol can be used to make urllib request a different host
        self.assertFalse(_validate_path("/@attacker.com/evil"))
        self.assertFalse(_validate_path("/path@host"))
    
    def test_validate_path_valid_paths(self):
        """Valid paths are accepted."""
        from cog.share import _validate_path
        
        self.assertTrue(_validate_path("/"))
        self.assertTrue(_validate_path("/graph/"))
        self.assertTrue(_validate_path("/graph/query"))
        self.assertTrue(_validate_path("/graph/stats"))
        self.assertTrue(_validate_path("/my_graph-123/"))
        self.assertTrue(_validate_path("/graph?param=value"))
        self.assertTrue(_validate_path("/graph?a=1&b=2"))


class TestTemplatesShareUrl(unittest.TestCase):
    """Test template URL generation with share_url."""
    
    def test_render_graph_row(self):
        """Graph row generates local href."""
        from cog.templates import render_graph_row
        
        html = render_graph_row("test_graph", 100, "rw")
        self.assertIn('href="/test_graph/"', html)
        self.assertIn(">test_graph<", html)
    
    def test_render_graph_row_escapes_xss(self):
        """Graph row escapes potentially malicious graph names."""
        from cog.templates import render_graph_row
        
        # Attempt XSS via graph name
        html = render_graph_row('<script>alert(1)</script>', 100, "rw")
        # Should be escaped, not executed
        self.assertNotIn("<script>", html)
        self.assertIn("&lt;script&gt;", html)
    
    def test_render_index_page_with_share_url(self):
        """Index page shows share URL when provided."""
        from cog.templates import render_index_page
        
        html = render_index_page(
            version="3.6.3",
            local_ip="127.0.0.1",
            port=8080,
            graphs_html="<div>test</div>",
            uptime_str="0h 0m 1s",
            share_url="https://xyz789.s.cogdb.io/"
        )
        
        # Share URL should appear in the output (escaped)
        self.assertIn("s.cogdb.io", html)
    
    def test_render_status_page_with_share_url(self):
        """Status page shows correct connect URL with share."""
        from cog.templates import render_status_page
        
        html = render_status_page(
            version="3.6.3",
            local_ip="127.0.0.1",
            port=8080,
            graph_name="my_graph",
            instance_id="abc123",
            node_count=50,
            edge_count=3,
            uptime_str="0h 0m 1s",
            queries_served=10,
            last_query_str="5s ago",
            mode_str="writable",
            share_url="https://session123.s.cogdb.io/"
        )
        
        # Back link should be root
        self.assertIn('href="/"', html)
        # Connect URL should use share URL
        self.assertIn("s.cogdb.io", html)


class TestServerHeaders(unittest.TestCase):
    """Test server respects X-Base-Path and X-Share-Url headers."""
    
    @classmethod
    def setUpClass(cls):
        if os.path.exists("/tmp/" + DIR_NAME):
            shutil.rmtree("/tmp/" + DIR_NAME)
        os.makedirs("/tmp/" + DIR_NAME, exist_ok=True)
        
        from cog.torque import Graph
        cls.g = Graph(graph_name="share_test", cog_home=DIR_NAME)
        cls.g.put("alice", "knows", "bob")
        
        cls.port = 18090
        cls.g.serve(port=cls.port)
        time.sleep(0.2)
    
    @classmethod
    def tearDownClass(cls):
        cls.g.stop()
        cls.g.close()
        shutil.rmtree("/tmp/" + DIR_NAME)
    
    def _request_with_headers(self, path, headers=None):
        """Helper to make HTTP requests with custom headers."""
        url = f"http://localhost:{self.port}{path}"
        req = urllib.request.Request(url)
        if headers:
            for key, value in headers.items():
                req.add_header(key, value)
        
        with urllib.request.urlopen(req, timeout=5) as response:
            return response.read().decode('utf-8')
    
    def test_index_page_with_share_url_header(self):
        """Index page shows share URL from header."""
        response = self._request_with_headers('/', headers={
            'X-Share-Url': 'https://test_session.s.cogdb.io/'
        })
        
        self.assertIn('s.cogdb.io', response)
    
    def test_status_page_with_headers(self):
        """Status page uses headers for connect URL."""
        response = self._request_with_headers('/share_test/', headers={
            'X-Share-Url': 'https://mysession.s.cogdb.io/'
        })
        
        # Connect URL should show share URL
        self.assertIn('s.cogdb.io', response)


class TestServeShareParameter(unittest.TestCase):
    """Test the share parameter on g.serve()."""
    
    def test_serve_sets_share_url_attribute(self):
        """serve(share=True) would set _share_url on successful connection."""
        # Note: We can't test actual relay connection without mocking
        # This test verifies the attribute exists and is initially None
        from cog.torque import Graph
        import shutil
        
        test_dir = "/tmp/ShareParamTest"
        if os.path.exists(test_dir):
            shutil.rmtree(test_dir)
        os.makedirs(test_dir, exist_ok=True)
        
        try:
            g = Graph(graph_name="param_test", cog_home="ShareParamTest")
            
            # Initially, _share_url should not exist
            self.assertFalse(hasattr(g, '_share_url'))
            
            # After serve without share, still shouldn't exist
            g.serve(port=18091)
            time.sleep(0.1)
            self.assertFalse(hasattr(g, '_share_url'))
            
            g.stop()
            g.close()
        finally:
            shutil.rmtree(test_dir)


if __name__ == '__main__':
    unittest.main()



"""
Test to verify that predicate tables are only created with hashed names,
not with raw predicate names (regression test for duplicate file bug).
"""
from cog.torque import Graph
from cog.database import hash_predicate
from cog import config as cfg
import unittest
import os
import shutil

DIR_NAME = "TestFileCount"


class TestPredicateFileCreation(unittest.TestCase):
    """
    Ensure Graph.put() and Graph.put_batch() only create tables with hashed
    predicate names, not raw predicate names.
    """

    @classmethod
    def setUpClass(cls):
        cls.base_path = os.path.join(cfg.COG_PATH_PREFIX, DIR_NAME)
        if os.path.exists(cls.base_path):
            shutil.rmtree(cls.base_path)
        os.makedirs(cls.base_path, exist_ok=True)

    def test_put_creates_only_hashed_predicate_files(self):
        """Verify Graph.put() doesn't create files with raw predicate names."""
        graph_name = "test_put_hashed"
        g = Graph(graph_name=graph_name, cog_home=DIR_NAME)
        
        # Use distinctive predicate names that would be obvious if stored raw
        predicates = ["KNOWS", "LIKES", "WORKS_AT"]
        
        g.put("alice", "KNOWS", "bob")
        g.put("alice", "LIKES", "charlie")
        g.put("bob", "WORKS_AT", "acme")
        
        # Get the actual graph directory from config
        graph_dir = g.config.cog_data_dir(graph_name)
        g.close()
        
        files = os.listdir(graph_dir)
        
        # No file should contain raw predicate names
        for predicate in predicates:
            raw_files = [f for f in files if f.startswith(predicate + "-")]
            self.assertEqual(
                len(raw_files), 0,
                f"Found files with raw predicate name '{predicate}': {raw_files}. "
                f"Predicates should be hashed before creating table files."
            )
        
        # Verify hashed predicate files DO exist
        for predicate in predicates:
            hashed = hash_predicate(predicate)
            hashed_files = [f for f in files if f.startswith(hashed + "-")]
            self.assertGreaterEqual(
                len(hashed_files), 1,
                f"No files found for hashed predicate {hashed} (from '{predicate}')"
            )

    def test_put_batch_creates_only_hashed_predicate_files(self):
        """Verify Graph.put_batch() doesn't create files with raw predicate names."""
        graph_name = "test_batch_hashed"
        g = Graph(graph_name=graph_name, cog_home=DIR_NAME)
        
        predicates = ["FOLLOWS", "MENTIONS", "REPLIES_TO"]
        
        g.put_batch([
            ("user1", "FOLLOWS", "user2"),
            ("user1", "MENTIONS", "user3"),
            ("user2", "REPLIES_TO", "user1"),
        ])
        
        # Get the actual graph directory from config
        graph_dir = g.config.cog_data_dir(graph_name)
        g.close()
        
        files = os.listdir(graph_dir)
        
        # No file should contain raw predicate names
        for predicate in predicates:
            raw_files = [f for f in files if f.startswith(predicate + "-")]
            self.assertEqual(
                len(raw_files), 0,
                f"Found files with raw predicate name '{predicate}': {raw_files}. "
                f"put_batch should use hashed predicate names."
            )
        
        # Verify hashed predicate files DO exist
        for predicate in predicates:
            hashed = hash_predicate(predicate)
            hashed_files = [f for f in files if f.startswith(hashed + "-")]
            self.assertGreaterEqual(
                len(hashed_files), 1,
                f"No files found for hashed predicate {hashed} (from '{predicate}')"
            )

    def test_expected_file_count(self):
        """Verify the expected number of files is created (no duplicates)."""
        graph_name = "test_file_count"
        g = Graph(graph_name=graph_name, cog_home=DIR_NAME)
        
        # 3 predicates
        g.put("a", "P1", "b")
        g.put("a", "P2", "c")
        g.put("a", "P3", "d")
        
        # Get the actual graph directory from config
        graph_dir = g.config.cog_data_dir(graph_name)
        g.close()
        
        files = os.listdir(graph_dir)
        
        # Expected: 3 predicate tables + 2 system tables = 5 tables
        # Each table = 2 files (index + store) = 10 files total
        # System tables: TOR_NODE_SET, TOR_EDGE_SET
        self.assertEqual(
            len(files), 10,
            f"Expected 10 files (3 predicates + 2 system tables × 2 files each), "
            f"but found {len(files)}: {sorted(files)}"
        )

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(cls.base_path):
            shutil.rmtree(cls.base_path)


if __name__ == '__main__':
    unittest.main()

